# flake8: noqa
# Copied from tag v0.1.0 https://github.com/NVIDIA/modulus-makani/tree/0218658e925a3c88c6513022a33ecc1647735c2e
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from functools import partial
from typing import Tuple

# for annotation of models
import torch
import torch.nn as nn

# get spectral transforms from torch_harmonics
import torch_harmonics as th
from torch.utils.checkpoint import checkpoint

from .layers import MLP, DropPath, EncoderDecoder, InverseRealFFT2, RealFFT2
from .spectral_convolution import FactorizedSpectralConv, SpectralConv


class SpectralFilterLayer(nn.Module):
    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        filter_type="linear",
        operator_type="diagonal",
        hidden_size_factor=1,
        factorization=None,
        rank=1.0,
        separable=False,
        complex_activation="real",
        spectral_layers=1,
        bias=False,
        drop_rate=0.0,
        gain=1.0,
    ):
        super(SpectralFilterLayer, self).__init__()

        if filter_type == "non-linear":
            raise NotImplementedError("legacy nonlinear complex filter is removed")

        elif filter_type == "linear" and factorization is None:
            self.filter = SpectralConv(
                forward_transform,
                inverse_transform,
                embed_dim,
                embed_dim,
                operator_type=operator_type,
                separable=separable,
                bias=bias,
                gain=gain,
            )

        elif filter_type == "linear" and factorization is not None:
            self.filter = FactorizedSpectralConv(
                forward_transform,
                inverse_transform,
                embed_dim,
                embed_dim,
                operator_type=operator_type,
                rank=rank,
                factorization=factorization,
                separable=separable,
                bias=bias,
                gain=gain,
            )

        else:
            raise (NotImplementedError)

    def forward(self, x):
        return self.filter(x)


class FourierNeuralOperatorBlock(nn.Module):
    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        filter_type="linear",
        operator_type="diagonal",
        mlp_ratio=2.0,
        mlp_drop_rate=0.0,
        path_drop_rate=0.0,
        act_layer=nn.GELU,
        norm_layer=(nn.Identity, nn.Identity),
        rank=1.0,
        factorization=None,
        separable=False,
        inner_skip="linear",
        outer_skip=None,
        use_mlp=False,
        comm_feature_inp_name=None,
        comm_feature_hidden_name=None,
        complex_activation="real",
        spectral_layers=1,
        bias=False,
        final_activation=False,
        checkpointing=0,
    ):
        super(FourierNeuralOperatorBlock, self).__init__()

        # determine some shapes
        self.input_shape_loc = (forward_transform.nlat, forward_transform.nlon)
        self.output_shape_loc = (inverse_transform.nlat, inverse_transform.nlon)

        # norm layer
        self.norm0 = norm_layer[0]()

        if act_layer == nn.Identity:
            gain_factor = 1.0
        else:
            gain_factor = 2.0

        if inner_skip == "linear":
            self.inner_skip = nn.Conv2d(embed_dim, embed_dim, 1, 1, bias=False)
            gain_factor /= 2.0
            nn.init.normal_(
                self.inner_skip.weight, std=math.sqrt(gain_factor / embed_dim)
            )
        elif inner_skip == "identity":
            self.inner_skip = nn.Identity()
            gain_factor /= 2.0
        elif inner_skip == "none":
            pass
        else:
            raise ValueError(f"Unknown skip connection type {inner_skip}")

        # convolution layer
        self.filter = SpectralFilterLayer(
            forward_transform,
            inverse_transform,
            embed_dim,
            filter_type,
            operator_type,
            hidden_size_factor=mlp_ratio,
            factorization=factorization,
            rank=rank,
            separable=separable,
            complex_activation=complex_activation,
            spectral_layers=spectral_layers,
            bias=bias,
            drop_rate=path_drop_rate,
            gain=gain_factor,
        )

        self.act_layer0 = act_layer()

        # norm layer
        self.norm1 = norm_layer[1]()

        if final_activation and act_layer != nn.Identity:
            gain_factor = 2.0
        else:
            gain_factor = 1.0

        if outer_skip == "linear":
            self.outer_skip = nn.Conv2d(embed_dim, embed_dim, 1, 1, bias=False)
            gain_factor /= 2.0
            torch.nn.init.normal_(
                self.outer_skip.weight, std=math.sqrt(gain_factor / embed_dim)
            )
        elif outer_skip == "identity":
            self.outer_skip = nn.Identity()
            gain_factor /= 2.0
        elif outer_skip == "none":
            pass
        else:
            raise ValueError(f"Unknown skip connection type {outer_skip}")

        if use_mlp:
            mlp_hidden_dim = int(embed_dim * mlp_ratio)
            self.mlp = MLP(
                in_features=embed_dim,
                hidden_features=mlp_hidden_dim,
                act_layer=act_layer,
                drop_rate=mlp_drop_rate,
                drop_type="features",
                comm_inp_name=comm_feature_inp_name,
                comm_hidden_name=comm_feature_hidden_name,
                checkpointing=checkpointing,
                gain=gain_factor,
            )

        # dropout
        self.drop_path = (
            DropPath(path_drop_rate) if path_drop_rate > 0.0 else nn.Identity()
        )

        if final_activation:
            self.act_layer1 = act_layer()

    def forward(self, x):
        """
        Updated FNO block
        """

        x, residual = self.filter(x)

        x = self.norm0(x)

        if hasattr(self, "inner_skip"):
            x = x + self.inner_skip(residual)

        if hasattr(self, "act_layer0"):
            x = self.act_layer0(x)

        if hasattr(self, "mlp"):
            x = self.mlp(x)

        x = self.norm1(x)

        x = self.drop_path(x)

        if hasattr(self, "outer_skip"):
            x = x + self.outer_skip(residual)

        if hasattr(self, "act_layer1"):
            x = self.act_layer1(x)

        return x


class SphericalFourierNeuralOperatorNet(nn.Module):
    """
    SFNO implementation as in Bonev et al.; Spherical Fourier Neural Operators: Learning Stable Dynamics on the Sphere
    """

    def __init__(
        self,
        spectral_transform="sht",
        model_grid_type="equiangular",
        sht_grid_type="legendre-gauss",
        filter_type="linear",
        operator_type="dhconv",
        inp_shape: Tuple[int, int] = (721, 1440),
        out_shape: Tuple[int, int] = (721, 1440),
        scale_factor=8,
        inp_chans=2,
        out_chans=2,
        embed_dim=32,
        num_layers=4,
        repeat_layers=1,
        use_mlp=True,
        mlp_ratio=2.0,
        encoder_ratio=1,
        decoder_ratio=1,
        activation_function="gelu",
        encoder_layers=1,
        pos_embed="none",
        pos_drop_rate=0.0,
        path_drop_rate=0.0,
        mlp_drop_rate=0.0,
        normalization_layer="instance_norm",
        max_modes=None,
        hard_thresholding_fraction=1.0,
        big_skip=True,
        rank=1.0,
        factorization=None,
        separable=False,
        complex_activation="real",
        spectral_layers=3,
        bias=False,
        checkpointing=0,
        **kwargs,
    ):
        super(SphericalFourierNeuralOperatorNet, self).__init__()

        self.inp_shape = inp_shape
        self.out_shape = out_shape
        self.inp_chans = inp_chans
        self.out_chans = out_chans
        self.embed_dim = embed_dim
        self.repeat_layers = repeat_layers
        self.big_skip = big_skip
        self.checkpointing = checkpointing

        # compute the downscaled image size
        self.h = int(self.inp_shape[0] // scale_factor)
        self.w = int(self.inp_shape[1] // scale_factor)

        # initialize spectral transforms
        self._init_spectral_transforms(
            spectral_transform,
            model_grid_type,
            sht_grid_type,
            hard_thresholding_fraction,
            max_modes,
        )

        # determine activation function
        if activation_function == "relu":
            activation_function = nn.ReLU
        elif activation_function == "gelu":
            activation_function = nn.GELU
        elif activation_function == "silu":
            activation_function = nn.SiLU
        else:
            raise ValueError(f"Unknown activation function {activation_function}")

        # set up encoder
        self.encoder = EncoderDecoder(
            num_layers=encoder_layers,
            input_dim=self.inp_chans,
            output_dim=self.embed_dim,
            hidden_dim=int(encoder_ratio * self.embed_dim),
            act_layer=activation_function,
            input_format="nchw",
        )
        fblock_mlp_inp_name = "fin"
        fblock_mlp_hidden_name = "fout"

        # dropout
        self.pos_drop = (
            nn.Dropout(p=pos_drop_rate) if pos_drop_rate > 0.0 else nn.Identity()
        )
        dpr = [x.item() for x in torch.linspace(0, path_drop_rate, num_layers)]

        # pick norm layer
        if normalization_layer == "layer_norm":
            raise NotImplementedError("requires makani distributed libraries")
        elif normalization_layer == "instance_norm":
            norm_layer_inp = partial(
                nn.InstanceNorm2d,
                num_features=embed_dim,
                eps=1e-6,
                affine=True,
                track_running_stats=False,
            )
            norm_layer_out = norm_layer_mid = norm_layer_inp
        elif normalization_layer == "none":
            norm_layer_out = norm_layer_mid = norm_layer_inp = nn.Identity
        else:
            raise NotImplementedError(
                f"Error, normalization {normalization_layer} not implemented."
            )

        # FNO blocks
        self.blocks = nn.ModuleList([])
        for i in range(num_layers):
            first_layer = i == 0
            last_layer = i == num_layers - 1

            forward_transform = self.trans_down if first_layer else self.trans
            inverse_transform = self.itrans_up if last_layer else self.itrans

            inner_skip = "none"
            outer_skip = "linear"

            if first_layer:
                norm_layer = (norm_layer_inp, norm_layer_mid)
            elif last_layer:
                norm_layer = (norm_layer_mid, norm_layer_out)
            else:
                norm_layer = (norm_layer_mid, norm_layer_mid)

            block = FourierNeuralOperatorBlock(
                forward_transform,
                inverse_transform,
                embed_dim,
                filter_type=filter_type,
                operator_type=operator_type,
                mlp_ratio=mlp_ratio,
                mlp_drop_rate=mlp_drop_rate,
                path_drop_rate=dpr[i],
                act_layer=activation_function,
                norm_layer=norm_layer,
                inner_skip=inner_skip,
                outer_skip=outer_skip,
                use_mlp=use_mlp,
                comm_feature_inp_name=fblock_mlp_inp_name,
                comm_feature_hidden_name=fblock_mlp_hidden_name,
                rank=rank,
                factorization=factorization,
                separable=separable,
                complex_activation=complex_activation,
                spectral_layers=spectral_layers,
                bias=bias,
                checkpointing=checkpointing,
            )

            self.blocks.append(block)

        self.decoder = EncoderDecoder(
            num_layers=encoder_layers,
            input_dim=embed_dim,
            output_dim=self.out_chans,
            hidden_dim=int(decoder_ratio * embed_dim),
            act_layer=activation_function,
            gain=0.5 if self.big_skip else 1.0,
            input_format="nchw",
        )

        # output transform
        if self.big_skip:
            self.residual_transform = nn.Conv2d(
                self.inp_chans, self.out_chans, 1, bias=False
            )
            self.residual_transform.weight.is_shared_mp = ["spatial"]
            self.residual_transform.weight.sharded_dims_mp = [None, None, None, None]
            scale = math.sqrt(0.5 / self.inp_chans)
            nn.init.normal_(self.residual_transform.weight, mean=0.0, std=scale)

        # learned position embedding
        if pos_embed == "direct":
            # currently using deliberately a differently shape position embedding
            self.pos_embed = nn.Parameter(
                torch.zeros(1, embed_dim, self.inp_shape_loc[0], self.inp_shape_loc[1])
            )
            # information about how tensors are shared / sharded across ranks
            self.pos_embed.is_shared_mp = []  # no reduction required since pos_embed is already serial
            self.pos_embed.sharded_dims_mp = [None, None, "h", "w"]
            self.pos_embed.type = "direct"
            with torch.no_grad():
                nn.init.trunc_normal_(self.pos_embed, std=0.02)
        elif pos_embed == "frequency":
            lmax_loc = self.itrans_up.lmax
            mmax_loc = self.itrans_up.mmax

            rcoeffs = nn.Parameter(
                torch.tril(torch.randn(1, embed_dim, lmax_loc, mmax_loc), diagonal=0)
            )
            ccoeffs = nn.Parameter(
                torch.tril(
                    torch.randn(1, embed_dim, lmax_loc, mmax_loc - 1), diagonal=-1
                )
            )
            with torch.no_grad():
                nn.init.trunc_normal_(rcoeffs, std=0.02)
                nn.init.trunc_normal_(ccoeffs, std=0.02)
            self.pos_embed = nn.ParameterList([rcoeffs, ccoeffs])
            self.pos_embed.type = "frequency"
            self.pos_embed.is_shared_mp = []
            self.pos_embed.sharded_dims_mp = [None, None, "h", "w"]

        elif pos_embed == "none" or pos_embed == "None" or pos_embed is None:
            pass
        else:
            raise ValueError("Unknown position embedding type")

    @torch.jit.ignore
    def _init_spectral_transforms(
        self,
        spectral_transform="sht",
        model_grid_type="equiangular",
        sht_grid_type="legendre-gauss",
        hard_thresholding_fraction=1.0,
        max_modes=None,
    ):
        """
        Initialize the spectral transforms based on the maximum number of modes to keep. Handles the computation
        of local image shapes and domain parallelism, based on the
        """

        if max_modes is not None:
            modes_lat, modes_lon = max_modes
        else:
            modes_lat = int(self.h * hard_thresholding_fraction)
            modes_lon = int((self.w // 2 + 1) * hard_thresholding_fraction)

        # prepare the spectral transforms
        if spectral_transform == "sht":
            sht_handle = th.RealSHT
            isht_handle = th.InverseRealSHT

            # set up
            self.trans_down = sht_handle(
                *self.inp_shape, lmax=modes_lat, mmax=modes_lon, grid=model_grid_type
            ).float()
            self.itrans_up = isht_handle(
                *self.out_shape, lmax=modes_lat, mmax=modes_lon, grid=model_grid_type
            ).float()
            self.trans = sht_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon, grid=sht_grid_type
            ).float()
            self.itrans = isht_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon, grid=sht_grid_type
            ).float()

        elif spectral_transform == "fft":
            fft_handle = RealFFT2
            ifft_handle = InverseRealFFT2

            self.trans_down = fft_handle(
                self.inp_shape[0], self.inp_shape[1], lmax=modes_lat, mmax=modes_lon
            ).float()
            self.itrans_up = ifft_handle(
                self.out_shape[0], self.out_shape[1], lmax=modes_lat, mmax=modes_lon
            ).float()
            self.trans = fft_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon
            ).float()
            self.itrans = ifft_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon
            ).float()
        else:
            raise (ValueError("Unknown spectral transform"))

        # use the SHT/FFT to compute the local, downscaled grid dimensions
        self.inp_shape_loc = (self.trans_down.nlat, self.trans_down.nlon)
        self.out_shape_loc = (self.itrans_up.nlat, self.itrans_up.nlon)
        self.h_loc = self.itrans.nlat
        self.w_loc = self.itrans.nlon

    @torch.jit.ignore
    def no_weight_decay(self):
        return {"pos_embed", "cls_token"}

    def _forward_features(self, x):
        for r in range(self.repeat_layers):
            for blk in self.blocks:
                if self.checkpointing >= 3:
                    x = checkpoint(blk, x, use_reentrant=False)
                else:
                    x = blk(x)

        return x

    def forward(self, x):
        # save big skip
        if self.big_skip:
            # if output shape differs, use the spectral transforms to change resolution
            if self.out_shape != self.inp_shape:
                xtype = x.dtype
                # only take the predicted channels as residual
                residual = x.to(torch.float32)
                with torch.amp.autocast("cuda", enabled=False):
                    residual = self.trans_down(residual)
                    residual = residual.contiguous()
                    residual = self.itrans_up(residual)
                    residual = residual.to(dtype=xtype)
            else:
                # only take the predicted channels
                residual = x

        if self.checkpointing >= 1:
            x = checkpoint(self.encoder, x, use_reentrant=False)
        else:
            x = self.encoder(x)

        if hasattr(self, "pos_embed"):
            if self.pos_embed.type == "frequency":
                pos_embed = torch.stack(
                    [
                        self.pos_embed[0],
                        nn.functional.pad(self.pos_embed[1], (1, 0), "constant", 0),
                    ],
                    dim=-1,
                )
                with torch.amp.autocast("cuda", enabled=False):
                    pos_embed = self.itrans_up(torch.view_as_complex(pos_embed))
            else:
                pos_embed = self.pos_embed

            # add pos embed
            x = x + pos_embed

        # maybe clean the padding just in case
        x = self.pos_drop(x)

        # do the feature extraction
        x = self._forward_features(x)

        if self.checkpointing >= 1:
            x = checkpoint(self.decoder, x, use_reentrant=False)
        else:
            x = self.decoder(x)

        if self.big_skip:
            x = x + self.residual_transform(residual)

        return x



import datetime
from contextlib import contextmanager

import numpy as np
import pytest
import torch
import xarray as xr

from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations

from .enso_coefficient import OVERLAP_THRESHOLD, EnsoCoefficientEvaluatorAggregator


@contextmanager
def change_aggregator_enso_index(
    aggregator: type[EnsoCoefficientEvaluatorAggregator], new_index: xr.DataArray
):
    """
    Temporarily change the class attribute enso index for testing purposes.
    """
    old_index = aggregator._enso_index
    aggregator._enso_index = new_index
    try:
        yield
    finally:
        aggregator._enso_index = old_index


def _data_generator(scale: int, steps: int):
    return torch.linspace(-scale, scale, steps)


def _get_data(
    scale: int,
    n_samples: int,
    n_times: int,
    n_lat: int,
    n_lon: int,
    calendar: str = "julian",
):
    time = xr.date_range(
        start="2000-01-01",
        periods=n_times,
        freq="6h",
        calendar=calendar,
        use_cftime=True,
    )
    enso_index = xr.DataArray(
        _data_generator(scale, n_times), dims=["time"], coords={"time": time}
    )
    target_data = {  # make target data perfectly correlated
        "a": torch.tile(
            _data_generator(scale, n_times)[None, :, None, None],
            [n_samples, 1, n_lat, n_lon],
        ).to(device=get_device()),
    }
    gen_data = {  # make gen data perfectly anti-correlated
        "a": torch.tile(
            _data_generator(-scale, n_times)[None, :, None, None],
            [n_samples, 1, n_lat, n_lon],
        ).to(device=get_device()),
    }
    sample_time = xr.concat(
        [
            xr.DataArray(
                time.values,
                dims=["time"],
            )
            for _ in range(n_samples)
        ],
        dim="sample",
    )
    return enso_index, sample_time, target_data, gen_data


@pytest.mark.parametrize("scaling", [0.5, 1.0, 2.0])
def test_enso_coefficient_aggregator_values(scaling):
    """
    Test the EnsoCoefficientEvaluatorAggregator class.

    Check that:
        - The aggregator maintains a zero-mean subset of the enso index for
            each sample.
        - The logged target and gen coefficients are scaled versions of 1.0 and
            -1.0 for perfectly correlated and anti-correlated data,respectively.
        - The logged global-mean coefficient RMSE is scaled from 2.0 (perfect
            correlation minus perfect anti-correlation).
        - The above two are also true via the `get_dataset` method.

    Args:
        scaling: How much to scale up or down the target and generated data;
            a scaling of 1.0 means that the target/gen data have the same scale
            as the enso index so the regression coefficients should be |1.0|.


    """
    n_samples, n_times, n_lat, n_lon = 2, 28, 3, 3
    scale = 3
    area_weights = torch.ones([n_lat, n_lon], device=get_device())
    # get data that doesn't vary in space, but varies in time with the ENSO index
    enso_index, sample_time, target_data, gen_data = _get_data(
        scale, n_samples, n_times, n_lat, n_lon
    )
    with change_aggregator_enso_index(EnsoCoefficientEvaluatorAggregator, enso_index):
        enso_agg = EnsoCoefficientEvaluatorAggregator(
            initial_time=sample_time.isel(time=0),
            n_forward_timesteps=(n_times - 1),
            timestep=datetime.timedelta(hours=6),
            gridded_operations=LatLonOperations(area_weights),
        )
    assert len(enso_agg._sample_index_series) == n_samples
    for index_values in enso_agg._sample_index_series:
        assert index_values is not None
        # check that the index values are zero-mean for each sample
        assert np.isclose(index_values.mean().item(), 0.0)
    target_data["a"] *= scaling
    gen_data["a"] *= scaling
    enso_agg.record_batch(time=sample_time, target_data=target_data, gen_data=gen_data)
    enso_agg.record_batch(time=sample_time, target_data=target_data, gen_data=gen_data)
    coefficients = enso_agg._get_coefficients()
    target_coefficients, gen_coefficients = coefficients
    assert target_coefficients is not None
    assert gen_coefficients is not None
    # check that the target coefficients are 1.0 * scaling (perfectly correlated)
    assert torch.allclose(target_coefficients["a"], torch.tensor(scaling))
    # check that the gen coefficients are -1.0 * scaling (perfectly anti-correlated)
    assert torch.allclose(gen_coefficients["a"], torch.tensor(-1.0 * scaling))
    logs = enso_agg.get_logs("enso_coefficients")
    # check that the global-mean coefficient RMSE is abs(-1.0 - 1.0)*scaling,
    # since the target and gen data are perfectly correlated and anti-correlated, resp.
    np.testing.assert_almost_equal(
        logs["enso_coefficients/rmse/a"], 2.0 * scaling, decimal=5
    )
    enso_dataset = enso_agg.get_dataset()
    # check that the coefficients are as expected in the dataset also
    np.testing.assert_array_almost_equal(
        enso_dataset["a"].sel(source="target").values,
        target_coefficients["a"].cpu().numpy(),
    )
    np.testing.assert_array_almost_equal(
        enso_dataset["a"].sel(source="prediction").values,
        gen_coefficients["a"].cpu().numpy(),
    )


@pytest.mark.parametrize("shift", [1.5, 0.95, 0.05, 0.0])
def test_enso_index_inference_overlap(shift):
    """
    Test the EnsoCoefficientEvaluatorAggregator returns metrics only for appropriate
    overlap of the reference index and the inference period.
    """
    n_samples, n_times, n_lat, n_lon = 2, 28, 3, 3
    data_scale = 3
    area_weights = torch.ones([n_lat, n_lon])
    enso_index, sample_time, target_data, gen_data = _get_data(
        data_scale, n_samples, n_times, n_lat, n_lon
    )
    # shift the sample times so they only partially overlap the reference index
    index_duration = enso_index.time[-1].item() - enso_index.time[0].item()
    offset_seconds = shift * index_duration.total_seconds()
    sample_time += datetime.timedelta(seconds=offset_seconds)
    with change_aggregator_enso_index(EnsoCoefficientEvaluatorAggregator, enso_index):
        enso_agg = EnsoCoefficientEvaluatorAggregator(
            initial_time=sample_time.isel(time=0),
            n_forward_timesteps=(n_times - 1),
            timestep=datetime.timedelta(hours=6),
            gridded_operations=LatLonOperations(area_weights),
        )
    enso_agg.record_batch(time=sample_time, target_data=target_data, gen_data=gen_data)
    target_coefficients, gen_coefficients = enso_agg._get_coefficients()
    overlap = max(1.0 - shift, 0.0)
    if overlap < OVERLAP_THRESHOLD:  # should be empty dict
        assert not target_coefficients
        assert not gen_coefficients
    else:
        assert target_coefficients is not None
        assert gen_coefficients is not None
        assert isinstance(target_coefficients["a"], torch.Tensor)
        assert isinstance(gen_coefficients["a"], torch.Tensor)


@pytest.mark.parametrize(
    "calendar",
    [
        pytest.param("julian", id="both_julian"),
        pytest.param("proleptic_gregorian", id="data_proleptic_gregorian_index_julian"),
    ],
)
def test_enso_agg_calendar(calendar):
    """
    Test the EnsoCoefficientEvaluatorAggregator class with different data calendars.

    The reference ENSO index is always in Julian calendar, but the data may be
    proleptic gregorian. The aggregator should be able to handle this.
    """
    n_samples, n_times, n_lat, n_lon = 2, 28, 3, 3
    data_scale = 3
    area_weights = torch.ones([n_lat, n_lon])
    enso_index, sample_time, target_data, gen_data = _get_data(
        data_scale, n_samples, n_times, n_lat, n_lon, calendar=calendar
    )
    enso_index = enso_index.convert_calendar("julian", dim="time", use_cftime=True)
    with change_aggregator_enso_index(EnsoCoefficientEvaluatorAggregator, enso_index):
        enso_agg = EnsoCoefficientEvaluatorAggregator(
            initial_time=sample_time.isel(time=0),
            n_forward_timesteps=(n_times - 1),
            timestep=datetime.timedelta(hours=6),
            gridded_operations=LatLonOperations(area_weights),
        )
    enso_agg.record_batch(time=sample_time, target_data=target_data, gen_data=gen_data)
    target_coefficients, gen_coefficients = enso_agg._get_coefficients()
    assert (target_coefficients is not None) and (gen_coefficients is not None)



import argparse
import os
from collections.abc import Sequence

import yaml

from .config import update_dict_with_dotlist


def prepare_config(path: str, override: Sequence[str] | None = None) -> dict:
    """Get config and update with possible dotlist override."""
    with open(path) as f:
        data = yaml.safe_load(f)
    data = update_dict_with_dotlist(data, override)
    return data


def prepare_directory(path: str, config_data: dict):
    """Create experiment directory and dump config_data to it."""
    if not os.path.isdir(path):
        os.makedirs(path, exist_ok=True)
    with open(os.path.join(path, "config.yaml"), "w") as f:
        yaml.dump(config_data, f, default_flow_style=False, sort_keys=False)


def get_parser():
    """Standard arg parser for ACE entrypoints."""
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml_config", type=str, help="Path to the YAML config file.")
    parser.add_argument(
        "--override",
        nargs="*",
        help="A dotlist of key=value pairs to override the config. "
        "For example, --override a.b=1 c=2, where a dot indicates nesting.",
    )
    return parser



# The dependencies of this script are installed in the "fv3net" conda environment
# which can be installed using fv3net's Makefile. See
# https://github.com/ai2cm/fv3net/blob/8ed295cf0b8ca49e24ae5d6dd00f57e8b30169ac/Makefile#L310

import dataclasses
import logging
import os
import shutil
import tempfile
import time
from typing import Dict, List, Literal, Optional

import click
import dacite
import fsspec
import xarray as xr
import yaml

# these are auxiliary variables that exist in dataset for convenience, e.g. to do
# masking or to more easily compute vertical integrals. But they are not inputs
# or outputs to the ML model, so we don't need normalization constants for them.
DROP_VARIABLES = (
    [
        "land_sea_mask",
        "pressure_thickness_of_atmospheric_layer_0",
        "pressure_thickness_of_atmospheric_layer_1",
        "pressure_thickness_of_atmospheric_layer_2",
        "pressure_thickness_of_atmospheric_layer_3",
        "pressure_thickness_of_atmospheric_layer_4",
        "pressure_thickness_of_atmospheric_layer_5",
        "pressure_thickness_of_atmospheric_layer_6",
        "pressure_thickness_of_atmospheric_layer_7",
    ]
    + [f"ak_{i}" for i in range(9)]
    + [f"bk_{i}" for i in range(9)]
)

DIMS = {
    "FV3GFS": ["time", "grid_xt", "grid_yt"],
    "E3SMV2": ["time", "lat", "lon"],
    "ERA5": ["time", "latitude", "longitude"],
    "CM4": ["time", "lat", "lon"],
}


def add_history_attrs(ds, input_zarr, start_date, end_date, n_samples):
    ds.attrs["history"] = (
        "Created by full-model/fv3gfs_data_process/get_stats.py. INPUT_ZARR:"
        f" {input_zarr}, START_DATE: {start_date}, END_DATE: {end_date}."
    )
    ds.attrs["input_samples"] = n_samples


def copy(source: str, destination: str):
    """Copy between any two 'filesystems'. Do not use for large files.

    Args:
        source: Path to source file/object.
        destination: Path to destination.
    """
    with fsspec.open(source) as f_source:
        with fsspec.open(destination, "wb") as f_destination:
            shutil.copyfileobj(f_source, f_destination)


@dataclasses.dataclass
class StatsConfig:
    output_directory: str
    data_type: Literal["FV3GFS", "E3SMV2", "ERA5", "CM4"]
    exclude_runs: List[str] = dataclasses.field(default_factory=list)
    start_date: Optional[str] = None
    end_date: Optional[str] = None
    beaker_dataset: Optional[str] = None


@dataclasses.dataclass
class Config:
    runs: Dict[str, str]
    data_output_directory: str
    stats: StatsConfig


@click.command()
@click.argument("config_yaml", type=str)
@click.argument("run", type=int)
@click.option(
    "--debug",
    is_flag=True,
    help="If set, print some statistics instead of writing normalization coefficients.",
)
def main(config_yaml: str, run: int, debug: bool):
    """
    Compute statistics for the data processing pipeline.

    Arguments:
    config_yaml -- Path to the configuration file for the data processing pipeline.
    run -- Run index for the data processing pipeline.
    """
    # Import dask-related things here to enable testing in environments without dask.
    import dask
    import distributed

    logging.basicConfig(level=logging.INFO)
    initial_time = time.time()
    distributed.Client(n_workers=16)

    with open(config_yaml, "r") as f:
        config_data = yaml.load(f, Loader=yaml.CLoader)
    config = dacite.from_dict(data_class=Config, data=config_data)
    xr.set_options(keep_attrs=True, display_max_rows=100)
    if config.data_output_directory.endswith("/"):
        config.data_output_directory = config.data_output_directory[:-1]
    if list(config.runs.keys())[run] in config.stats.exclude_runs:
        logging.info(f"Skipping run {list(config.runs.keys())[run]}")
        return
    input_zarr = (
        config.data_output_directory + "/" + list(config.runs.keys())[run] + ".zarr"
    )
    logging.info(f"Reading data from {input_zarr}")

    # Open data with roughly 128 MiB chunks via dask's automatic chunking. This
    # is useful when opening sharded zarr stores with an inner chunk size of 1,
    # which is otherwise inefficient for the type of computation done here.
    with dask.config.set({"array.chunk-size": "128MiB"}):
        ds = xr.open_zarr(input_zarr, chunks={"time": "auto"})

    ds = ds.drop_vars(DROP_VARIABLES, errors="ignore")
    ds = ds.sel(time=slice(config.stats.start_date, config.stats.end_date))

    dims = DIMS[config.stats.data_type]

    # Explicitly compute the statistics here, since xarray does not support
    # writing netCDFs with the scipy engine with the distributed scheduler.
    # There is no harm to computing here versus later, since the end result is
    # not something memory intensive.
    centering = ds.mean(dim=dims).compute()
    logging.info("Computed centering")
    scaling_full_field = ds.std(dim=dims).compute()
    logging.info("Computed scaling_full_field")
    scaling_residual = ds.diff("time").std(dim=dims).compute()
    logging.info("Computed scaling_resiudal")
    time_means = ds.mean(dim="time").compute()
    logging.info("Computed time_means")

    for dataset in [
        centering,
        scaling_full_field,
        scaling_residual,
        time_means,
    ]:
        n_samples = len(ds.time)
        add_history_attrs(
            dataset,
            input_zarr,
            config.stats.start_date,
            config.stats.end_date,
            n_samples,
        )

    if debug:
        normed_data = (ds - centering) / scaling_full_field
        logging.info(f"Average of normed data: {normed_data.mean(dim=dims).compute()}")
        logging.info(
            f"Standard deviation of normed data: {normed_data.std(dim=dims).compute()}"
        )
        all_var_stddev = normed_data.to_array().std(dim=["variable"] + dims)
        logging.info(
            f"Standard deviation computed over all variables: {all_var_stddev.values}"
        )
    else:
        out_dir = config.stats.output_directory + "/" + list(config.runs.keys())[run]
        if out_dir.startswith("gs:"):
            temp_dir = tempfile.TemporaryDirectory()
            local_dir = temp_dir.name
            remote_dir: Optional[str] = out_dir
        else:
            if not os.path.isdir(out_dir):
                os.makedirs(out_dir)
            local_dir = out_dir
            remote_dir = None

        centering.to_netcdf(os.path.join(local_dir, "centering.nc"))
        if remote_dir is not None:
            copy(
                os.path.join(local_dir, "centering.nc"),
                remote_dir + "/centering.nc",
            )
        scaling_full_field.to_netcdf(os.path.join(local_dir, "scaling-full-field.nc"))
        if remote_dir is not None:
            copy(
                os.path.join(local_dir, "scaling-full-field.nc"),
                remote_dir + "/scaling-full-field.nc",
            )
        scaling_residual.to_netcdf(os.path.join(local_dir, "scaling-residual.nc"))
        if remote_dir is not None:
            copy(
                os.path.join(local_dir, "scaling-residual.nc"),
                remote_dir + "/scaling-residual.nc",
            )
        time_means.to_netcdf(os.path.join(local_dir, "time-mean.nc"))
        if remote_dir is not None:
            copy(
                os.path.join(local_dir, "time-mean.nc"),
                remote_dir + "/time-mean.nc",
            )

    total_time = time.time() - initial_time
    logging.info(f"Total time for computing stats: {total_time:0.2f} seconds.")


if __name__ == "__main__":
    main()



import datetime
import json
import logging
import os
import re
import warnings
from collections import namedtuple
from functools import lru_cache
from urllib.parse import urlparse

import fsspec
import numpy as np
import torch
import xarray as xr
from xarray.coding.times import CFDatetimeCoder

from fme.core.coordinates import (
    DepthCoordinate,
    HorizontalCoordinates,
    HybridSigmaPressureCoordinate,
    NullVerticalCoordinate,
    VerticalCoordinate,
)
from fme.core.dataset.properties import DatasetProperties
from fme.core.mask_provider import MaskProvider
from fme.core.stacker import Stacker
from fme.core.typing_ import TensorDict

from .config import XarrayDataConfig
from .data_typing import VariableMetadata
from .utils import (
    as_broadcasted_tensor,
    get_horizontal_coordinates,
    get_nonspacetime_dimensions,
    load_series_data,
    load_series_data_zarr_async,
)

SLICE_NONE = slice(None)
logger = logging.getLogger(__name__)

VariableNames = namedtuple(
    "VariableNames",
    (
        "time_dependent_names",
        "time_invariant_names",
        "static_derived_names",
    ),
)


def _get_vertical_coordinate(
    ds: xr.Dataset, dtype: torch.dtype | None
) -> VerticalCoordinate:
    """
    Get vertical coordinate from a dataset.

    If the dataset contains variables named `ak_N` and `bk_N` where
    `N` is the level number, then a hybrid sigma-pressure coordinate
    will be returned. If the dataset contains variables named
    `idepth_N` then a depth coordinate will be returned. If neither thing
    is true, a hybrid sigma-pressure coordinate of lenght 0 is returned.

    Args:
        ds: Dataset to get vertical coordinates from.
        dtype: Data type of the returned tensors. If None, the dtype is not
            changed from the original in ds.
    """
    ak_mapping = {
        int(v[3:]): torch.as_tensor(ds[v].values)
        for v in ds.variables
        if v.startswith("ak_")
    }
    bk_mapping = {
        int(v[3:]): torch.as_tensor(ds[v].values)
        for v in ds.variables
        if v.startswith("bk_")
    }
    ak_list = [ak_mapping[k] for k in sorted(ak_mapping.keys())]
    bk_list = [bk_mapping[k] for k in sorted(bk_mapping.keys())]

    idepth_mapping = {
        int(v[7:]): torch.as_tensor(ds[v].values)
        for v in ds.variables
        if v.startswith("idepth_")
    }
    idepth_list = [idepth_mapping[k] for k in sorted(idepth_mapping.keys())]

    if len(ak_list) > 0 and len(bk_list) > 0 and len(idepth_list) > 0:
        raise ValueError(
            "Dataset contains both hybrid sigma-pressure and depth coordinates. "
            "Can only provide one, or else the vertical coordinate is ambiguous."
        )

    coordinate: VerticalCoordinate
    surface_mask = None
    if len(idepth_list) > 0:
        if "mask_0" in ds.data_vars:
            mask_layers = {
                name: torch.as_tensor(ds[name].values, dtype=dtype)
                for name in ds.data_vars
                if re.match(r"mask_(\d+)$", name)
            }
            for name in mask_layers:
                if "time" in ds[name].dims:
                    raise ValueError("The ocean mask must by time-independent.")
            stacker = Stacker({"mask": ["mask_"]})
            mask = stacker("mask", mask_layers)
            if "surface_mask" in ds.data_vars:
                if "time" in ds["surface_mask"].dims:
                    raise ValueError("The surface mask must be time-independent.")
                surface_mask = torch.as_tensor(ds["surface_mask"].values, dtype=dtype)
        else:
            logger.warning(
                "Dataset does not contain a mask. Providing a DepthCoordinate with "
                "mask set to 1 at all layers."
            )
            mask = torch.ones(len(idepth_list) - 1, dtype=dtype)
        coordinate = DepthCoordinate(
            torch.as_tensor(idepth_list, dtype=dtype), mask, surface_mask
        )
    elif len(ak_list) > 0 and len(bk_list) > 0:
        coordinate = HybridSigmaPressureCoordinate(
            ak=torch.as_tensor(ak_list, dtype=dtype),
            bk=torch.as_tensor(bk_list, dtype=dtype),
        )
    else:
        logger.warning("Dataset does not contain a vertical coordinate.")
        coordinate = NullVerticalCoordinate()

    return coordinate


def _get_raw_times(paths: list[str], engine: str) -> list[np.ndarray]:
    times = []
    for path in paths:
        with _open_xr_dataset(path, engine=engine) as ds:
            times.append(ds.time.values)
    return times


def _repeat_and_increment_time(
    raw_times: list[np.ndarray], n_repeats: int, timestep: datetime.timedelta
) -> list[np.ndarray]:
    """Repeats and increments a collection of arrays of evenly spaced times."""
    n_timesteps = sum(len(times) for times in raw_times)
    timespan = timestep * n_timesteps

    repeated_and_incremented_time = []
    for repeats in range(n_repeats):
        increment = repeats * timespan
        for time in raw_times:
            incremented_time = time + increment
            repeated_and_incremented_time.append(incremented_time)
    return repeated_and_incremented_time


def _get_cumulative_timesteps(time: list[np.ndarray]) -> np.ndarray:
    """Returns a list of cumulative timesteps for each item in a time coordinate."""
    num_timesteps_per_file = [0]
    for time_coord in time:
        num_timesteps_per_file.append(len(time_coord))
    return np.array(num_timesteps_per_file).cumsum()


def _get_file_local_index(index: int, start_indices: np.ndarray) -> tuple[int, int]:
    """
    Return a tuple of the index of the file containing the time point at `index`
    and the index of the time point within that file.
    """
    file_index = np.searchsorted(start_indices, index, side="right") - 1
    time_index = index - start_indices[file_index]
    return int(file_index), time_index


class StaticDerivedData:
    names = ("x", "y", "z")
    metadata = {
        "x": VariableMetadata(units="", long_name="Euclidean x-coordinate"),
        "y": VariableMetadata(units="", long_name="Euclidean y-coordinate"),
        "z": VariableMetadata(units="", long_name="Euclidean z-coordinate"),
    }

    def __init__(self, coordinates: HorizontalCoordinates):
        self._coords = coordinates
        self._x: torch.Tensor | None = None
        self._y: torch.Tensor | None = None
        self._z: torch.Tensor | None = None

    def _get_xyz(self) -> TensorDict:
        if self._x is None or self._y is None or self._z is None:
            coords = self._coords
            x, y, z = coords.xyz

            self._x = torch.as_tensor(x)
            self._y = torch.as_tensor(y)
            self._z = torch.as_tensor(z)

        return {"x": self._x, "y": self._y, "z": self._z}

    def __getitem__(self, name: str) -> torch.Tensor:
        return self._get_xyz()[name]


def _get_protocol(path):
    return urlparse(str(path)).scheme


def _get_fs(path):
    protocol = _get_protocol(path)
    if not protocol:
        protocol = "file"
    proto_kw = _get_fs_protocol_kwargs(path)
    fs = fsspec.filesystem(protocol, **proto_kw)

    return fs


def _preserve_protocol(original_path, glob_paths):
    protocol = _get_protocol(str(original_path))
    if protocol:
        glob_paths = [f"{protocol}://{path}" for path in glob_paths]
    return glob_paths


def _get_fs_protocol_kwargs(path):
    protocol = _get_protocol(path)
    kwargs = {}
    if protocol == "gs":
        # https://gcsfs.readthedocs.io/en/latest/api.html#gcsfs.core.GCSFileSystem
        key_json = os.environ.get("FSSPEC_GS_KEY_JSON", None)
        key_file = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None)

        if key_json is not None:
            token = json.loads(key_json)
        elif key_file is not None:
            token = key_file
        else:
            logger.warning(
                "GCS currently expects user credentials authenticated using"
                " `gcloud auth application-default login`. This is not recommended for "
                "production use."
            )
            token = "google_default"
        kwargs["token"] = token
    elif protocol == "s3":
        # https://s3fs.readthedocs.io/en/latest/#s3-compatible-storage
        env_vars = [
            "FSSPEC_S3_KEY",
            "FSSPEC_S3_SECRET",
            "FSSPEC_S3_ENDPOINT_URL",
        ]
        for v in env_vars:
            if v not in os.environ:
                warnings.warn(
                    f"An S3 path was specified but environment variable {v} "
                    "was not found. This may cause authentication issues if not "
                    "set and no other defaults are present. See "
                    "https://s3fs.readthedocs.io/en/latest/#s3-compatible-storage"
                    " for details."
                )

    return kwargs


def _open_xr_dataset(path: str, *args, **kwargs):
    # need the path to get protocol specific arguments for the backend
    protocol_kw = _get_fs_protocol_kwargs(path)
    if protocol_kw:
        kwargs.update({"storage_options": protocol_kw})

    return xr.open_dataset(
        path,
        *args,
        decode_times=CFDatetimeCoder(use_cftime=True),
        decode_timedelta=False,
        mask_and_scale=False,
        cache=False,
        chunks=None,
        **kwargs,
    )


_open_xr_dataset_lru = lru_cache()(_open_xr_dataset)


def _open_file_fh_cached(path, **kwargs):
    protocol = _get_protocol(path)
    if protocol:
        # add an LRU cache for remote zarrs
        return _open_xr_dataset_lru(
            path,
            **kwargs,
        )
    # netcdf4 and h5engine have a filehandle LRU cache in xarray
    # https://github.com/pydata/xarray/blob/cd3ab8d5580eeb3639d38e1e884d2d9838ef6aa1/xarray/backends/file_manager.py#L54 # noqa: E501
    return _open_xr_dataset(
        path,
        **kwargs,
    )


def get_raw_paths(path, file_pattern):
    fs = _get_fs(path)
    glob_paths = sorted(fs.glob(os.path.join(path, file_pattern)))
    raw_paths = _preserve_protocol(path, glob_paths)
    return raw_paths


def _get_mask_provider(ds: xr.Dataset, dtype: torch.dtype | None) -> MaskProvider:
    """
    Get mask provider from a dataset.

    If the dataset contains static variables that start with the string "mask_" or a
    variable named "surface_mask", then these variables will be used to instantiate
    a MaskProvider object. Otherwise, an empty MaskProvider is returned.

    Args:
        ds: Dataset to get vertical coordinates from.
        dtype: Data type of the returned tensors. If None, the dtype is not
            changed from the original in ds.
    """
    masks: dict[str, torch.Tensor] = {
        name: torch.as_tensor(ds[name].values, dtype=dtype)
        for name in ds.data_vars
        if "mask_" in name
    }
    for name in masks:
        if "time" in ds[name].dims:
            raise ValueError("Masks must be time-independent.")
    mask_provider = MaskProvider(masks)
    logging.info(f"Initialized {mask_provider}.")
    return mask_provider


class XarrayDataset(torch.utils.data.Dataset):
    """Load data from a directory of files matching a pattern using xarray. The
    number of contiguous timesteps to load for each sample is specified by the
    n_timesteps argument.

    For example, if the file(s) have the time coordinate
    (t0, t1, t2, t3, t4) and n_timesteps=3, then this dataset will
    provide three samples: (t0, t1, t2), (t1, t2, t3), and (t2, t3, t4).
    """

    def __init__(self, config: XarrayDataConfig, names: list[str], n_timesteps: int):
        self._horizontal_coordinates: HorizontalCoordinates
        self._names = names
        self.path = config.data_path
        self.file_pattern = config.file_pattern
        self.engine = config.engine
        self.dtype = config.torch_dtype
        self.spatial_dimensions = config.spatial_dimensions
        self.fill_nans = config.fill_nans
        self.subset_config = config.subset
        self._raw_paths = get_raw_paths(self.path, self.file_pattern)
        if len(self._raw_paths) == 0:
            raise ValueError(
                f"No files found matching '{self.path}/{self.file_pattern}'."
            )
        self.full_paths = self._raw_paths * config.n_repeats
        self.sample_n_times = n_timesteps
        self._get_files_stats(config.n_repeats, config.infer_timestep)
        first_dataset = xr.open_dataset(
            self.full_paths[0],
            decode_times=False,
            decode_timedelta=False,
            engine=self.engine,
            chunks=None,
        )
        self._mask_provider = _get_mask_provider(first_dataset, self.dtype)
        (
            self._horizontal_coordinates,
            self._static_derived_data,
            _loaded_horizontal_dims,
        ) = self.configure_horizontal_coordinates(first_dataset)
        (
            self._time_dependent_names,
            self._time_invariant_names,
            self._static_derived_names,
        ) = self._group_variable_names_by_time_type()

        self._vertical_coordinate = _get_vertical_coordinate(first_dataset, self.dtype)
        self.overwrite = config.overwrite

        self._nonspacetime_dims = get_nonspacetime_dimensions(
            first_dataset, _loaded_horizontal_dims
        )
        self._shape_excluding_time = [
            first_dataset.sizes[dim]
            for dim in (self._nonspacetime_dims + _loaded_horizontal_dims)
        ]
        self._loaded_dims = ["time"] + self._nonspacetime_dims + _loaded_horizontal_dims
        self.isel = {
            dim: v if isinstance(v, int) else v.slice for dim, v in config.isel.items()
        }
        self._isel_tuple = tuple(
            [self.isel.get(dim, SLICE_NONE) for dim in self._loaded_dims[1:]]
        )
        self._check_isel_dimensions(first_dataset.sizes)

    def _check_isel_dimensions(self, data_dim_sizes):
        # Horizontal dimensions are not currently supported, as the current isel code
        # does not adjust HorizonalCoordinates to match selection.
        if "time" in self.isel:
            raise ValueError("isel cannot be used to select time. Use subset instead.")

        for dim, selection in self.isel.items():
            if dim not in self._nonspacetime_dims:
                raise ValueError(
                    f"isel dimension {dim} must be a non-spacetime dimension "
                    f"of the dataset ({self._nonspacetime_dims})."
                )
            max_isel_index = (
                (selection.start or 0) if isinstance(selection, slice) else selection
            )
            if max_isel_index >= data_dim_sizes[dim]:
                raise ValueError(
                    f"isel index {max_isel_index} is out of bounds for dimension "
                    f"{dim} with size {data_dim_sizes[dim]}."
                )

    @property
    def _shape_excluding_time_after_selection(self):
        final_shape = []
        for orig_size, sel in zip(self._shape_excluding_time, self._isel_tuple):
            # if selecting a single index, dimension is squeezed
            # so it is not included in the final shape
            if isinstance(sel, slice):
                if sel.start is None and sel.stop is None and sel.step is None:
                    final_shape.append(orig_size)
                else:
                    final_shape.append(len(range(*sel.indices(orig_size))))
        return final_shape

    @property
    def dims(self) -> list[str]:
        # Final dimensions of returned data after dims that are selected
        # with a single index are dropped
        final_dims = ["time"]
        for dim, sel in zip(self._loaded_dims[1:], self._isel_tuple):
            if isinstance(sel, slice):
                final_dims.append(dim)
        return final_dims

    @property
    def properties(self) -> DatasetProperties:
        return DatasetProperties(
            self._variable_metadata,
            self._vertical_coordinate,
            self._horizontal_coordinates,
            self._mask_provider,
            self.timestep,
            self._is_remote,
        )

    @property
    def _is_remote(self) -> bool:
        protocol = _get_protocol(str(self.path))
        if not protocol or protocol == "file":
            return False
        return True

    @property
    def all_times(self) -> xr.CFTimeIndex:
        """Time index of all available times in the data."""
        return self._all_times

    def _get_variable_metadata(self, ds):
        result = {}
        for name in self._names:
            if name in StaticDerivedData.names:
                result[name] = StaticDerivedData.metadata[name]
            elif hasattr(ds[name], "units") and hasattr(ds[name], "long_name"):
                result[name] = VariableMetadata(
                    units=ds[name].units,
                    long_name=ds[name].long_name,
                )
        self._variable_metadata = result

    def _get_files_stats(self, n_repeats: int, infer_timestep: bool):
        logging.info(f"Opening data at {os.path.join(self.path, self.file_pattern)}")
        raw_times = _get_raw_times(self._raw_paths, engine=self.engine)

        self._timestep: datetime.timedelta | None
        if infer_timestep:
            self._timestep = _get_timestep(np.concatenate(raw_times))
            time_coord = _repeat_and_increment_time(raw_times, n_repeats, self.timestep)
        else:
            self._timestep = None
            time_coord = raw_times

        cum_num_timesteps = _get_cumulative_timesteps(time_coord)
        self.start_indices = cum_num_timesteps[:-1]
        self.total_timesteps = cum_num_timesteps[-1]
        self._n_initial_conditions = self.total_timesteps - self.sample_n_times + 1
        self._sample_start_times = xr.CFTimeIndex(
            np.concatenate(time_coord)[: self._n_initial_conditions]
        )
        self._all_times = xr.CFTimeIndex(np.concatenate(time_coord))

        del cum_num_timesteps, time_coord

        ds = self._open_file(0)
        self._get_variable_metadata(ds)

        logging.info(f"Found {self._n_initial_conditions} samples.")

    def _group_variable_names_by_time_type(self) -> VariableNames:
        """Returns lists of time-dependent variable names, time-independent
        variable names, and variables which are only present as an initial
        condition.
        """
        (
            time_dependent_names,
            time_invariant_names,
            static_derived_names,
        ) = ([], [], [])
        # Don't use open_mfdataset here, because it will give time-invariant
        # fields a time dimension. We assume that all fields are present in the
        # netcdf file corresponding to the first chunk of time.
        with _open_xr_dataset(self.full_paths[0], engine=self.engine) as ds:
            for name in self._names:
                if name in StaticDerivedData.names:
                    static_derived_names.append(name)
                else:
                    try:
                        da = ds[name]
                    except KeyError:
                        raise ValueError(
                            f"Required variable not found in dataset: {name}."
                        )
                    else:
                        dims = da.dims
                        if "time" in dims:
                            time_dependent_names.append(name)
                        else:
                            time_invariant_names.append(name)
            logging.info(
                f"The required variables have been found in the dataset: {self._names}."
            )

        return VariableNames(
            time_dependent_names,
            time_invariant_names,
            static_derived_names,
        )

    def configure_horizontal_coordinates(
        self, first_dataset
    ) -> tuple[HorizontalCoordinates, StaticDerivedData, list[str]]:
        horizontal_coordinates: HorizontalCoordinates
        static_derived_data: StaticDerivedData

        horizontal_coordinates, dim_names = get_horizontal_coordinates(
            first_dataset, self.spatial_dimensions, self.dtype
        )
        static_derived_data = StaticDerivedData(horizontal_coordinates)

        coords_sizes = {
            coord_name: len(coord)
            for coord_name, coord in horizontal_coordinates.coords.items()
        }
        logging.info(f"Horizontal coordinate sizes are {coords_sizes}.")
        return horizontal_coordinates, static_derived_data, dim_names

    @property
    def timestep(self) -> datetime.timedelta:
        if self._timestep is None:
            raise ValueError(
                "Timestep was not inferred in the data loader. Note "
                "XarrayDataConfig.infer_timestep must be set to True for this "
                "to occur."
            )
        else:
            return self._timestep

    def __len__(self):
        return self._n_initial_conditions

    def _open_file(self, idx):
        logger.debug(f"Opening file {self.full_paths[idx]}")
        return _open_file_fh_cached(self.full_paths[idx], engine=self.engine)

    @property
    def sample_start_times(self) -> xr.CFTimeIndex:
        """Return cftime index corresponding to start time of each sample."""
        return self._sample_start_times

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        """Return a sample of data spanning the timesteps
        [idx, idx + self.sample_n_times).

        Args:
            idx: Index of the sample to retrieve.

        Returns:
            Tuple of a sample's data (i.e. a mapping from names to torch.Tensors) and
            its corresponding time coordinate.
        """
        time_slice = slice(idx, idx + self.sample_n_times)
        return self.get_sample_by_time_slice(time_slice)

    def get_sample_by_time_slice(
        self, time_slice: slice
    ) -> tuple[TensorDict, xr.DataArray]:
        input_file_idx, input_local_idx = _get_file_local_index(
            time_slice.start, self.start_indices
        )
        output_file_idx, output_local_idx = _get_file_local_index(
            time_slice.stop - 1, self.start_indices
        )

        # get the sequence of observations
        arrays: dict[str, list[torch.Tensor]] = {}
        idxs = range(input_file_idx, output_file_idx + 1)
        total_steps = 0
        for i, file_idx in enumerate(idxs):
            start = input_local_idx if i == 0 else 0
            if i == len(idxs) - 1:
                stop = output_local_idx
            else:
                stop = (
                    self.start_indices[file_idx + 1] - self.start_indices[file_idx] - 1
                )

            n_steps = stop - start + 1
            shape = [n_steps] + self._shape_excluding_time_after_selection
            total_steps += n_steps
            if self.engine == "zarr":
                tensor_dict = load_series_data_zarr_async(
                    idx=start,
                    n_steps=n_steps,
                    path=self.full_paths[file_idx],
                    names=self._time_dependent_names,
                    final_dims=self.dims,
                    final_shape=shape,
                    fill_nans=self.fill_nans,
                    nontime_selection=self._isel_tuple,
                )
            else:
                ds = self._open_file(file_idx)
                ds = ds.isel(**self.isel)
                tensor_dict = load_series_data(
                    idx=start,
                    n_steps=n_steps,
                    ds=ds,
                    names=self._time_dependent_names,
                    final_dims=self.dims,
                    final_shape=shape,
                    fill_nans=self.fill_nans,
                )
                ds.close()
                del ds
            for n in self._time_dependent_names:
                arrays.setdefault(n, []).append(tensor_dict[n])

        tensors: TensorDict = {}
        for n, tensor_list in arrays.items():
            tensors[n] = torch.cat(tensor_list)
        del arrays

        # load time-invariant variables from first dataset
        if len(self._time_invariant_names) > 0:
            ds = self._open_file(idxs[0])
            ds = ds.isel(**self.isel)
            shape = [total_steps] + self._shape_excluding_time_after_selection
            for name in self._time_invariant_names:
                variable = ds[name].variable
                if self.fill_nans is not None:
                    variable = variable.fillna(self.fill_nans.value)
                tensors[name] = as_broadcasted_tensor(variable, self.dims, shape)
            ds.close()
            del ds

        # load static derived variables
        for name in self._static_derived_names:
            tensor = self._static_derived_data[name]
            horizontal_dims = [1] * tensor.ndim
            tensors[name] = tensor.repeat((total_steps, *horizontal_dims))

        # cast to desired dtype
        tensors = {k: v.to(dtype=self.dtype) for k, v in tensors.items()}

        # Apply field overwrites
        tensors = self.overwrite.apply(tensors)

        # Create a DataArray of times to return corresponding to the slice that
        # is valid even when n_repeats > 1.
        time = xr.DataArray(self.all_times[time_slice].values, dims=["time"])

        return tensors, time


def _get_timestep(time: np.ndarray) -> datetime.timedelta:
    """Computes the timestep of an array of a time coordinate array.

    Raises an error if the times are not separated by a positive constant
    interval, or if the array has one or fewer times.
    """
    assert len(time.shape) == 1, "times must be a 1D array"

    if len(time) > 1:
        timesteps = np.diff(time)
        timestep = timesteps[0]

        if not (timestep > datetime.timedelta(days=0)):
            raise ValueError("Timestep of data must be greater than zero.")

        if not np.all(timesteps == timestep):
            raise ValueError("Time coordinate does not have a uniform timestep.")

        return timestep
    else:
        raise ValueError(
            "Time coordinate does not have enough times to infer a timestep."
        )



import re

import pytest
from compute_dataset import validate_vertical_coarsening_indices


@pytest.mark.parametrize(
    ("size", "indices", "valid"),
    [
        pytest.param(5, [(0, 2), (2, 5)], True, id="valid"),
        pytest.param(5, [(0, 2), (1, 5)], False, id="invalid-overlapping"),
        pytest.param(5, [(0, 2), (2, 4)], False, id="invalid-incomplete"),
        pytest.param(5, [(0, 2), (2, 6)], False, id="invalid-out-of-bounds"),
    ],
)
def test_validate_vertical_coarsening_indices(size, indices, valid):
    component = "atmosphere"
    control_flag = "validate_vertical_coarsening_indices"
    if valid:
        validate_vertical_coarsening_indices(size, indices, component, control_flag)
    else:
        # Check that both the component and control flag appear in the error message.
        match = re.compile(rf"(?=.*\b{component}\b)(?=.*\b{control_flag}\b)")
        with pytest.raises(ValueError, match=match):
            validate_vertical_coarsening_indices(size, indices, component, control_flag)



import os

from .cli import prepare_config, prepare_directory


def test_prepare_config(tmp_path):
    path = tmp_path / "file.yaml"
    with open(path, "w") as f:
        f.write("a: 2")
    output = prepare_config(path)
    assert output == {"a": 2}


def test_prepare_config_override(tmp_path):
    path = tmp_path / "file.yaml"
    with open(path, "w") as f:
        f.write("a: 2")
    output = prepare_config(path, ["a=3"])
    assert output == {"a": 3}


def test_prepare_directory(tmp_path):
    path = tmp_path / "test"
    config_data = {"a": 2}
    prepare_directory(path, config_data)
    assert os.path.exists(path / "config.yaml")



from collections.abc import Callable, Mapping
from typing import Any, Generic, TypeVar

import dacite

T = TypeVar("T")


class Registry(Generic[T]):
    """
    Used to register and initialize multiple types of a dataclass.
    """

    def __init__(self):
        """
        Initialize the registry.

        Args:
            default_type: if given, the "type" key in the config dict is optional
                and by default this type will be used.
        """
        self._types: dict[str, type[T]] = {}

    def register(self, type_name: str) -> Callable[[type[T]], type[T]]:
        """
        Registers a configuration type with the registry.

        When registry.get is called to initialize a dataclass, if the
        "type" argument passed is equal to the type_name you give here,
        then the decorated class will be the one initialized from the data
        in the "config" argument.

        Args:
            type_name: name used in configuration to indicate the decorated
                class as the target type to be initialized when using `get`.
        """

        def register_func(cls: type[T]) -> type[T]:
            base_type = None
            # attribute exists only when type parameter is passed
            # i.e. Registry[BaseClass]()
            if hasattr(self, "__orig_class__"):
                base_type = self.__orig_class__.__args__[0]

            if base_type and not issubclass(cls, base_type):
                raise TypeError(f"{cls} must be a subclass of {base_type}")

            self._types[type_name] = cls
            return cls

        return register_func

    def get(self, type_name: str, config: Mapping[str, Any]) -> T:
        return dacite.from_dict(
            data_class=self._types[type_name],
            data=config,
            config=dacite.Config(strict=True),
        )



"""
Exponential Moving Average (EMA) module.

Copied from https://github.com/CompVis/latent-diffusion/blob/main/ldm/modules/ema.py
and modified.

MIT License

Copyright (c) 2022 Machine Vision and Learning Group, LMU Munich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import dataclasses
from collections.abc import Iterable, Iterator
from typing import Protocol

import torch
from torch import nn


class HasNamedParameters(Protocol):
    def named_parameters(
        self, recurse: bool = True
    ) -> Iterator[tuple[str, nn.Parameter]]: ...


@dataclasses.dataclass
class EMAConfig:
    """
    Configuration for exponential moving average of model weights.

    Parameters:
        decay: decay rate for the moving average
    """

    decay: float = 0.9999

    def build(self, model: HasNamedParameters):
        return EMATracker(model, decay=self.decay, faster_decay_at_start=True)


class EMATracker:
    """
    Exponential Moving Average (EMA) tracker.

    This tracks the moving average of the parameters of a model, and has methods
    that can be used to temporarily replace the parameters of the model with its EMA.
    """

    def __init__(
        self, model: HasNamedParameters, decay: float, faster_decay_at_start=True
    ):
        """
        Create a new EMA tracker.

        Args:
            model: The model whose parameters should be tracked.
            decay: The decay rate of the moving average.
            faster_decay_at_start: Whether to use the number of updates to determine
                the decay rate. If True, the decay rate will be min(decay, (1 +
                num_updates) / (10 + num_updates)). If False, the decay rate
                will be decay.
        """
        super().__init__()
        if decay < 0.0 or decay > 1.0:
            raise ValueError("Decay must be between 0 and 1")

        self._module_name_to_ema_name = {}
        self.decay = torch.tensor(decay, dtype=torch.float32)
        self._faster_decay_at_start = faster_decay_at_start
        self.num_updates = torch.tensor(0, dtype=torch.int)

        self._ema_params = {}

        for name, p in model.named_parameters():
            if p.requires_grad:
                # remove as '.'-character is not allowed in buffers
                ema_name = name.replace(".", "")
                self._module_name_to_ema_name.update({name: ema_name})
                self._ema_params[ema_name] = p.clone().detach().data

        self._stored_params: list[nn.Parameter] = []

    def __call__(self, model: HasNamedParameters):
        """
        Update the moving average of the parameters.

        Does not mutate the input, only updates the moving average.

        Args:
            model: The model whose parameters should be updated. Should be a model
                specified identically to the one passed when this object was
                instantiated.
        """
        decay = self.decay

        self.num_updates += 1
        if self._faster_decay_at_start:
            decay = min(self.decay, (1 + self.num_updates) / (10 + self.num_updates))

        with torch.no_grad():
            module_parameters = dict(model.named_parameters())

            for key in module_parameters:
                if module_parameters[key].requires_grad:
                    ema_name = self._module_name_to_ema_name[key]
                    self._ema_params[ema_name] = self._ema_params[ema_name].type_as(
                        module_parameters[key]
                    )
                    self._ema_params[ema_name].sub_(
                        (1.0 - decay)
                        * (self._ema_params[ema_name] - module_parameters[key])
                    )
                elif key in self._module_name_to_ema_name:
                    raise ValueError(
                        f"Expected model parameter {key} to require gradient, "
                        "but it does not"
                    )

    def copy_to(self, model: HasNamedParameters):
        """
        Copy the averaged parameters to the model, overwriting its values.
        """
        m_param = dict(model.named_parameters())
        for key in m_param:
            if m_param[key].requires_grad:
                m_param[key].data.copy_(
                    self._ema_params[self._module_name_to_ema_name[key]].data
                )
            else:
                assert key not in self._module_name_to_ema_name

    def store(self, parameters: Iterable[nn.Parameter]):
        """
        Save the current parameters for restoring later.

        Args:
            parameters: The parameters to be stored for later restoration by `restore`
        """
        self._stored_params = [param.clone() for param in parameters]

    def restore(self, parameters: Iterable[nn.Parameter]):
        """
        Restore the parameters stored with the `store` method.

        Useful to validate the model with EMA parameters without affecting the
        original optimization process. Store the parameters before the
        `copy_to` method. After validation (or model saving), use this to
        restore the former parameters.

        Args:
            parameters: The parameters to be updated with the values stored by `store`
        """
        for c_param, param in zip(self._stored_params, parameters):
            param.data.copy_(c_param.data)

    def get_state(self):
        """
        Get the state of the EMA tracker, excluding weights.

        Returns:
            The state of the EMA tracker.
        """
        return {
            "decay": self.decay,
            "num_updates": self.num_updates,
            "faster_decay_at_start": self._faster_decay_at_start,
            "module_name_to_ema_name": self._module_name_to_ema_name,
        }

    @classmethod
    def from_state(cls, state, model) -> "EMATracker":
        """
        Create an EMA tracker from a state.

        Args:
            state: The state of the EMA tracker.
            model: The model whose parameters should be tracked, used to
                initialize the EMA weights. Should come from an EMA checkpoint.

        Returns:
            The EMA tracker.
        """
        ema = cls(model, float(state["decay"]), state["faster_decay_at_start"])
        ema.num_updates = state["num_updates"]
        ema._module_name_to_ema_name = state["module_name_to_ema_name"]
        return ema



import abc
import dataclasses
import re

import torch

from fme.ace.data_loading.batch_data import BatchData


@dataclasses.dataclass
class AugmentationConfig:
    """
    Configuration for data augmentation.

    Attributes:
        rotate_probability: The probability of rotating the sphere by 180 degrees,
            as a value between 0.0 and 1.0.
        additional_directional_names: Names of variables whose sign is flipped when
            the poles are reversed. By default this includes known directional
            names as stored in RotateModifier.FLIP_NAMES.
    """

    rotate_probability: float = 0.0
    additional_directional_names: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        if not 0.0 <= self.rotate_probability <= 1.0:
            raise ValueError(
                "rotate_probability must be between 0.0 and 1.0, "
                f"got {self.rotate_probability}"
            )

    def build_modifier(self) -> "BatchModifierABC":
        if self.rotate_probability > 0.0:
            return RotateModifier(
                self.rotate_probability, self.additional_directional_names
            )
        return NullModifier()


class BatchModifierABC(abc.ABC):
    @abc.abstractmethod
    def __call__(self, batch: BatchData) -> BatchData: ...


class RotateModifier(BatchModifierABC):
    """
    Modifier that rotates the sphere by 180 degrees so that the poles swap
    places. This is the same as flipping both zonal and meridional axes.

    Also flips the sign of horizontal directional variables such as horizontal
    winds in specific directions, so their new values reflect the rotated axes.
    The names of such variables are stored in the `FLIP_NAMES` class variable.
    Variables not included in this list are not flipped.

    Specifically, the regex pattern r'{name}(_?[0-9]+m?)?$' is used to match the
    names of variables whose sign is flipped when the poles are reversed, for
    each name in `FLIP_NAMES`. This will match both names that end with something
    like "_0", "_1", etc. or something like "10m" or "2m".

    Note that seasons are handled by the fact that solar insolation is a data
    variable, but time is not modified. This means monthly or seasonal averages
    using this data will be affected by the rotation.
    """

    # names of variables whose sign is flipped when the poles are reversed
    FLIP_NAMES = [
        "eastward_wind",
        "northward_wind",
        "UGRD",
        "VGRD",
        "U",
        "V",
    ]

    def __init__(
        self,
        rotate_probability: float,
        additional_directional_names: list[str],
    ):
        self.rotate_probability = rotate_probability
        self.additional_directional_names = additional_directional_names
        self._pattern = re.compile(
            r"({})(_?[0-9]+m?)?$".format(
                "|".join(self.FLIP_NAMES + self.additional_directional_names)
            )
        )

    def __call__(self, batch: BatchData) -> BatchData:
        if batch.horizontal_dims != ["lat", "lon"]:
            raise NotImplementedError(
                "Horizontal dimensions must be lat and lon to rotate the sphere, got "
                f"{batch.horizontal_dims}"
            )
        example_value = next(iter(batch.data.values()))
        apply = torch.rand(example_value.shape[0]) < self.rotate_probability
        while len(apply.shape) < len(example_value.shape):
            apply = apply.unsqueeze(-1)
        new_data = {}
        for name, value in batch.data.items():
            new_value = torch.flip(value, dims=[-2, -1])
            if self._pattern.match(name):
                new_value = -1 * new_value
            new_data[name] = torch.where(apply, new_value, value)
        return BatchData(new_data, batch.time, batch.horizontal_dims)


class NullModifier(BatchModifierABC):
    def __call__(self, batch: BatchData) -> BatchData:
        return batch



from collections.abc import Callable, Mapping
from types import MappingProxyType
from typing import Protocol

import torch

from fme.core.constants import DENSITY_OF_WATER_CM4, SPECIFIC_HEAT_OF_WATER_CM4
from fme.core.stacker import Stacker
from fme.core.typing_ import TensorDict, TensorMapping

OCEAN_FIELD_NAME_PREFIXES = MappingProxyType(
    {
        "sea_water_potential_temperature": ["thetao_"],
        "sea_water_salinity": ["so_"],
        "sea_water_x_velocity": ["uo_"],
        "sea_water_y_velocity": ["vo_"],
        "sea_surface_height_above_geoid": ["zos"],
        "sea_surface_temperature": ["sst"],
        "sea_ice_fraction": ["sea_ice_fraction"],
        "sea_ice_thickness": ["sea_ice_thickness"],
        "sea_ice_volume": ["sea_ice_volume"],
        "ocean_sea_ice_fraction": ["ocean_sea_ice_fraction"],
        "land_fraction": ["land_fraction"],
        "net_downward_surface_heat_flux": ["hfds"],
        "geothermal_heat_flux": ["hfgeou"],
        "sea_surface_fraction": ["sea_surface_fraction"],
    }
)


class HasOceanDepthIntegral(Protocol):
    def __len__(self) -> int: ...

    def get_mask(self) -> torch.Tensor: ...

    def get_mask_level(self, level: int) -> torch.Tensor: ...

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None: ...

    def get_idepth(self) -> torch.Tensor: ...

    def depth_integral(
        self,
        integrand: torch.Tensor,
    ) -> torch.Tensor: ...

    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]: ...

    def to(self, device: str) -> "HasOceanDepthIntegral": ...


class OceanData:
    """Container for ocean data for accessing variables and providing
    torch.Tensor views on data with multiple depth levels.
    """

    def __init__(
        self,
        ocean_data: TensorMapping,
        depth_coordinate: HasOceanDepthIntegral | None = None,
        ocean_field_name_prefixes: Mapping[str, list[str]] = OCEAN_FIELD_NAME_PREFIXES,
    ):
        """
        Initializes the instance based on the provided data and prefixes.

        Args:
            ocean_data: Mapping from field names to tensors.
            depth_coordinate: The depth coordinate of the model.
            ocean_field_name_prefixes: Mapping which defines the correspondence
                between an arbitrary set of "standard" names (e.g.,
                "potential_temperature" or "salinity") and lists of possible
                names or prefix variants (e.g., ["thetao_"] or
                ["zos"]) found in the data.
        """
        self._data = dict(ocean_data)
        self._prefix_map = ocean_field_name_prefixes
        self._depth_coordinate = depth_coordinate
        self._stacker = Stacker(ocean_field_name_prefixes)

    @property
    def data(self) -> TensorDict:
        """Mapping from field names to tensors."""
        return self._data

    def __getitem__(self, name: str):
        return getattr(self, name)

    def _get_prefix(self, prefix):
        return self.data[prefix]

    def _set(self, name, value):
        for prefix in self._prefix_map[name]:
            if prefix in self.data.keys():
                self._set_prefix(prefix, value)
                return
        raise KeyError(name)

    def _set_prefix(self, prefix, value):
        self.data[prefix] = value

    def _get(self, name):
        for prefix in self._prefix_map[name]:
            if prefix in self.data.keys():
                return self._get_prefix(prefix)
        raise KeyError(name)

    @property
    def sea_water_potential_temperature(self) -> torch.Tensor:
        """Returns all depth levels of potential temperature."""
        return self._stacker("sea_water_potential_temperature", self.data)

    @property
    def sea_water_salinity(self) -> torch.Tensor:
        """Returns all depth levels of salinity."""
        return self._stacker("sea_water_salinity", self.data)

    @property
    def sea_water_x_velocity(self) -> torch.Tensor:
        """Returns all depth levels of x-velocity."""
        return self._stacker("sea_water_x_velocity", self.data)

    @property
    def sea_water_y_velocity(self) -> torch.Tensor:
        """Returns all depth levels of y-velocity."""
        return self._stacker("sea_water_y_velocity", self.data)

    @property
    def sea_surface_temperature(self) -> torch.Tensor:
        """Returns surface temperature."""
        return self._get("sea_surface_temperature")

    @property
    def sea_surface_height_above_geoid(self) -> torch.Tensor:
        """Returns sea surface height above geoid."""
        return self._get("sea_surface_height_above_geoid")

    @property
    def ocean_heat_content(self) -> torch.Tensor:
        """Returns column-integrated ocean heat content."""
        if self._depth_coordinate is None:
            raise ValueError(
                "Depth coordinate must be provided to compute column-integrated "
                "ocean heat content."
            )
        return self._depth_coordinate.depth_integral(
            self.sea_water_potential_temperature
            * SPECIFIC_HEAT_OF_WATER_CM4
            * DENSITY_OF_WATER_CM4
        )

    @property
    def sea_surface_fraction(self) -> torch.Tensor:
        """Returns the sea surface fraction."""
        return self._get("sea_surface_fraction")

    @property
    def net_downward_surface_heat_flux(self) -> torch.Tensor:
        """Net heat flux downward across the ocean surface (below the sea-ice)."""
        return self._get("net_downward_surface_heat_flux")

    @property
    def geothermal_heat_flux(self) -> torch.Tensor:
        """Geothermal heat flux."""
        return self._get("geothermal_heat_flux")

    @property
    def sea_ice_fraction(self) -> torch.Tensor:
        """Returns the sea ice fraction."""
        try:
            return self._get("sea_ice_fraction")
        except KeyError:
            land_fraction = self.land_fraction
            ocean_sea_ice_fraction = self.ocean_sea_ice_fraction
            return ocean_sea_ice_fraction * (1 - land_fraction)

    @property
    def land_fraction(self) -> torch.Tensor:
        """Returns the land fraction."""
        return self._get("land_fraction")

    @property
    def ocean_sea_ice_fraction(self) -> torch.Tensor:
        """Returns the sea ice fraction as a proportion of the sea surface."""
        return self._get("ocean_sea_ice_fraction")

    @property
    def ocean_fraction(self) -> torch.Tensor:
        """Returns the dynamic ocean fraction, computed from the sea ice
        fraction and land fraction.
        """
        return 1 - self.land_fraction - self.sea_ice_fraction



import os

import dacite
import pytest
import yaml
from combine_stats import Config as CombineStatsConfig
from get_stats import Config as GetStatsConfig
from upload_stats import Config as UploadStatsConfig

DIRNAME = os.path.abspath(os.path.dirname(__file__))
# list files in DIRNAME/config
CONFIG_YAMLS = [
    os.path.join(DIRNAME + "/configs", f)
    for f in os.listdir(DIRNAME + "/configs")
    if f.endswith(".yaml")
]


@pytest.mark.parametrize(
    "filename",
    CONFIG_YAMLS,
)
@pytest.mark.parametrize("cls", [GetStatsConfig, UploadStatsConfig, CombineStatsConfig])
def test_get_stats_valid(filename, cls):
    with open(filename, "r") as f:
        config_data = yaml.load(f, Loader=yaml.CLoader)
    dacite.from_dict(data_class=cls, data=config_data)



import dataclasses
import datetime
import pathlib
from collections.abc import Iterable, Sequence
from typing import Any

import cftime
import numpy as np
import xarray as xr

from fme.ace.data_loading.inference import (
    InferenceDataLoaderConfig,
    InferenceInitialConditionIndices,
)
from fme.core.coordinates import DimSize
from fme.core.dataset.config import XarrayDataConfig


def _coord_value(
    name, size, timestep: datetime.timedelta = datetime.timedelta(hours=6)
):
    # xarray data loader requires time to be a cftime.datetime object
    if name == "time":
        return [
            cftime.DatetimeProlepticGregorian(2000, 1, 1) + i * timestep
            for i in range(size)
        ]
    else:
        return np.arange(size, dtype=np.float32)


@dataclasses.dataclass
class DimSizes:
    n_time: int
    horizontal: list[DimSize]
    nz_interface: int

    @property
    def shape_nd(self) -> list[int]:
        return [self.n_time] + [dim.size for dim in self.horizontal]

    @property
    def dims_nd(self) -> list[str]:
        return ["time"] + [dim.name for dim in self.horizontal]

    @property
    def shape_vertical_interface(self) -> tuple[int]:
        return (self.nz_interface,)

    @property
    def items(self):
        return [("time", self.n_time)] + [
            (dim.name, dim.size) for dim in self.horizontal
        ]

    def get_size(self, key: str) -> int:
        for item in self.horizontal:
            if item.name == key:
                return item.size
        raise KeyError(f"Dimension with name '{key}' not found.")


def save_nd_netcdf(
    filename,
    dim_sizes: DimSizes,
    variable_names: list[str],
    timestep_days: float,
    time_varying_values: list[float] | None = None,
    save_vertical_coordinate: bool = True,
    return_ds: bool = False,
) -> xr.Dataset | None:
    """
    Save a ND netcdf file with random data for the given variable names and
    dimensions.

    Args:
        filename: The filename to save the netcdf file to.
        dim_sizes: The dimensions of the data.
        variable_names: The names of the variables to save.
        timestep_days: The number of days between each time step.
        time_varying_values: If not None, the values to use for each time step.
        save_vertical_coordinate: If True, save vertical coordinate variables.
        return_ds: If True, return the dataset in addition to saving it.
    """
    ds = get_nd_dataset(
        dim_sizes=dim_sizes,
        variable_names=variable_names,
        timestep_days=timestep_days,
        include_vertical_coordinate=save_vertical_coordinate,
    )
    if time_varying_values is not None:
        for name in variable_names:
            for i in range(dim_sizes.n_time):
                ds[name].isel(time=i).values[:] = time_varying_values[i]
    ds.to_netcdf(filename, unlimited_dims=["time"], format="NETCDF4_CLASSIC")
    if return_ds:
        return ds
    return None


def save_scalar_netcdf(
    filename,
    variable_names: list[str],
):
    ds = get_scalar_dataset(variable_names)
    ds.to_netcdf(filename, format="NETCDF4_CLASSIC")


@dataclasses.dataclass
class StatsData:
    path: pathlib.Path
    names: list[str]

    def __post_init__(self):
        save_scalar_netcdf(
            self.mean_filename,
            variable_names=self.names,
        )
        save_scalar_netcdf(
            self.std_filename,
            variable_names=self.names,
        )

    @property
    def mean_filename(self):
        return self.path / "stats-mean.nc"

    @property
    def std_filename(self):
        return self.path / "stats-std.nc"


@dataclasses.dataclass
class FV3GFSData:
    path: pathlib.Path
    names: list[str]
    dim_sizes: DimSizes
    timestep_days: float
    time_varying_values: list[float] | None = None
    num_data_workers: int = 0
    save_vertical_coordinate: bool = True

    def __post_init__(self):
        self.data_path.mkdir(parents=True, exist_ok=True)
        if (
            self.time_varying_values is not None
            and len(self.time_varying_values) != self.dim_sizes.n_time
        ):
            raise ValueError(
                f"Number of time-varying values ({len(self.time_varying_values)}) "
                f"must match number of time steps ({self.dim_sizes.n_time})"
            )
        self._ds: xr.Dataset = save_nd_netcdf(
            self.data_filename,
            dim_sizes=self.dim_sizes,
            variable_names=self.names,
            timestep_days=self.timestep_days,
            time_varying_values=self.time_varying_values,
            save_vertical_coordinate=self.save_vertical_coordinate,
            return_ds=True,
        )

    @property
    def data_path(self):
        # data must be in a separate path as a default loader loads all *.nc files
        return self.path / "data"

    @property
    def data_filename(self):
        return self.data_path / "data.nc"

    @property
    def horizontal_coords(self) -> dict[str, xr.DataArray]:
        return {dim.name: self._ds[dim.name] for dim in self.dim_sizes.horizontal}

    @property
    def inference_data_loader_config(self) -> InferenceDataLoaderConfig:
        return InferenceDataLoaderConfig(
            dataset=XarrayDataConfig(
                str(self.data_path),
            ),
            start_indices=InferenceInitialConditionIndices(
                first=0, n_initial_conditions=1, interval=1
            ),
            num_data_workers=self.num_data_workers,
        )


@dataclasses.dataclass
class MonthlyReferenceData:
    path: pathlib.Path
    names: list[str]
    dim_sizes: DimSizes
    n_ensemble: int

    def __post_init__(self):
        self.data_path.mkdir(parents=True, exist_ok=True)
        ds = get_nd_dataset(
            dim_sizes=self.dim_sizes,
            variable_names=self.names,
        )
        # drop default time coord from get_nd_dataset
        ds = ds.drop_vars(["time"])
        # add a time axis for months
        months = []
        for i in range(self.dim_sizes.n_time):
            year = 2000 + i // 12
            month = i % 12 + 1
            months.append(cftime.DatetimeProlepticGregorian(year, month, 15))
        ds["counts"] = xr.DataArray(
            np.random.randint(1, 10, size=(self.dim_sizes.n_time,)),
            dims=["time"],
            attrs={"units": "m", "long_name": "counts", "coordinates": "valid_time"},
        )
        member_datasets = []
        months_list = []
        for _ in range(self.n_ensemble):
            member_datasets.append(ds)
            months_list.append(xr.DataArray(months, dims=["time"]))
        ds = xr.concat(member_datasets, dim="sample")
        ds.coords["valid_time"] = xr.concat(months_list, dim="sample")
        ds.to_netcdf(self.data_filename, format="NETCDF4_CLASSIC")
        self.start_time = cftime.DatetimeProlepticGregorian(2000, 1, 1)

    @property
    def data_path(self):
        # data must be in a separate path as a default loader loads all *.nc files
        return self.path / "monthly_data"

    @property
    def data_filename(self):
        return self.data_path / "monthly.nc"


def get_nd_overrides(dim_sizes: DimSizes, timestep_days: float) -> dict[str, Any]:
    coords_override: dict[str, Any]
    time = [
        cftime.DatetimeProlepticGregorian(2000, 1, 1)
        + i * timestep_days * datetime.timedelta(days=1)
        for i in range(dim_sizes.n_time)
    ]
    coords_override = {"time": time}
    if "grid_yt" in dim_sizes.dims_nd:
        n_lat = dim_sizes.get_size("grid_yt")  # n_lat
        n_lon = dim_sizes.get_size("grid_xt")  # n_lon
        grid_yt = np.linspace(-89.5, 89.5, n_lat)
        grid_xt_start = 360.0 / n_lon / 2
        grid_xt = np.linspace(grid_xt_start, 360.0 - grid_xt_start, n_lon)
        coords_override["grid_yt"] = grid_yt
        coords_override["grid_xt"] = grid_xt
    return coords_override


def get_nd_dataset(
    dim_sizes: DimSizes,
    variable_names: Sequence[str],
    timestep_days: float = 1.0,
    include_vertical_coordinate: bool = True,
):
    """
    Gets a dataset of [time, <horizontal dims>] data.
    """
    data_vars = {}
    for name in variable_names:
        data = np.random.randn(*dim_sizes.shape_nd).astype(np.float32)
        data_vars[name] = xr.DataArray(
            data,
            dims=dim_sizes.dims_nd,
            attrs={"units": "m", "long_name": name},
        )
    coords_override = get_nd_overrides(dim_sizes=dim_sizes, timestep_days=timestep_days)

    coords = {
        dim_name: (
            xr.DataArray(
                np.arange(size, dtype=np.float64),
                dims=(dim_name,),
            )
            if dim_name not in coords_override
            else coords_override[dim_name]
        )
        for dim_name, size in dim_sizes.items
    }

    if include_vertical_coordinate:
        for i in range(dim_sizes.nz_interface):
            data_vars[f"ak_{i}"] = np.float64(i)
            data_vars[f"bk_{i}"] = np.float64(i + 1)

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    return ds


def get_scalar_dataset(
    variable_names: Iterable[str],
    fill_value: float | None = None,
):
    data_vars = {}
    for name in variable_names:
        if fill_value is not None:
            value = fill_value
        else:
            value = np.random.randn()
        data_vars[name] = xr.DataArray(value, attrs={"units": "m", "long_name": name})

    ds = xr.Dataset(data_vars=data_vars)
    return ds



import numpy as np
import pytest

from fme.core.wandb import DirectInitializationError, Image, WandB


def test_image_is_image_instance():
    wandb = WandB.get_instance()
    img = wandb.Image(np.zeros((10, 10)))
    assert isinstance(img, Image)


def test_wandb_direct_initialization_raises():
    with pytest.raises(DirectInitializationError):
        Image(np.zeros((10, 10)))



import abc
import dataclasses
import math
import re
from collections.abc import Callable, Mapping
from datetime import timedelta
from typing import Literal, TypeVar

import dacite
import numpy as np
import torch

try:
    from earth2grid import healpix as e2ghpx
except ImportError:
    e2ghpx = None

from fme.core import metrics
from fme.core.constants import GRAVITY
from fme.core.corrector.atmosphere import AtmosphereCorrector, AtmosphereCorrectorConfig
from fme.core.corrector.ocean import OceanCorrector, OceanCorrectorConfig
from fme.core.corrector.registry import CorrectorABC
from fme.core.derived_variables import compute_derived_quantities
from fme.core.device import get_device
from fme.core.gridded_ops import GriddedOperations, HEALPixOperations, LatLonOperations
from fme.core.mask_provider import MaskProvider, MaskProviderABC, NullMaskProvider
from fme.core.masking import StaticMasking
from fme.core.ocean_derived_variables import compute_ocean_derived_quantities
from fme.core.registry.corrector import CorrectorSelector
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.winds import lon_lat_to_xyz

HC = TypeVar("HC", bound="HorizontalCoordinates")


class DeriveFnABC(abc.ABC):
    @abc.abstractmethod
    def __call__(self, data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        pass


class NullDeriveFn(DeriveFnABC):
    def __call__(self, data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        return dict(data)


class AtmosphericDeriveFn(DeriveFnABC):
    def __init__(
        self,
        vertical_coordinate: "OptionalHybridSigmaPressureCoordinate",
        timestep: timedelta,
    ):
        self.vertical_coordinate = vertical_coordinate.to(
            "cpu"
        )  # must be on cpu for multiprocessing fork context
        self.timestep = timestep

    def __call__(self, data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        if isinstance(self.vertical_coordinate, NullVerticalCoordinate):
            vertical_coord: HybridSigmaPressureCoordinate | None = None
        else:
            vertical_coord = self.vertical_coordinate.to(get_device())
        return compute_derived_quantities(
            dict(data),
            vertical_coordinate=vertical_coord,
            timestep=self.timestep,
            forcing_data=dict(forcing_data),
        )


class OceanDeriveFn(DeriveFnABC):
    def __init__(
        self,
        depth_coordinate: "OptionalDepthCoordinate",
        timestep: timedelta,
    ):
        self.depth_coordinate = depth_coordinate.to(
            "cpu"
        )  # must be on cpu for multiprocessing fork context
        self.timestep = timestep

    def __call__(self, data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        if isinstance(self.depth_coordinate, NullVerticalCoordinate):
            depth_coord: DepthCoordinate | None = None
        else:
            depth_coord = self.depth_coordinate.to(get_device())
        return compute_ocean_derived_quantities(
            dict(data),
            depth_coordinate=depth_coord,
            timestep=self.timestep,
            forcing_data=dict(forcing_data),
        )


PostProcessFnType = Callable[[TensorMapping], TensorDict]


class NullPostProcessFn:
    def __call__(self, data: TensorMapping) -> TensorDict:
        return dict(data)


class VerticalCoordinate(abc.ABC):
    """
    A vertical coordinate system for use in the stepper.
    """

    SelfType = TypeVar("SelfType", bound="VerticalCoordinate")

    @abc.abstractmethod
    def to(self: SelfType, device: str) -> SelfType:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abc.abstractmethod
    def build_corrector(
        self,
        config: AtmosphereCorrectorConfig | CorrectorSelector,
        gridded_operations: GriddedOperations,
        timestep: timedelta,
    ) -> CorrectorABC:
        pass

    @abc.abstractmethod
    def build_derive_function(self, timestep: timedelta) -> DeriveFnABC:
        pass

    @property
    @abc.abstractmethod
    def coords(self) -> dict[str, np.ndarray]:
        pass

    @abc.abstractmethod
    def as_dict(self) -> TensorMapping:
        pass


@dataclasses.dataclass
class HybridSigmaPressureCoordinate(VerticalCoordinate):
    """
    Defines pressure at interface levels according to the following formula:
        p(k) = a(k) + b(k)*ps.

    where ps is the surface pressure, a and b are the sigma-pressure coordinates.

    Parameters:
        ak: a(k) coefficients as a 1-dimensional tensor
        bk: b(k) coefficients as a 1-dimensional tensor
    """

    ak: torch.Tensor
    bk: torch.Tensor

    def __post_init__(self):
        if len(self.ak.shape) != 1:
            raise ValueError(
                f"ak must be a 1-dimensional tensor. Got shape: {self.ak.shape}"
            )
        if len(self.bk.shape) != 1:
            raise ValueError(
                f"bk must be a 1-dimensional tensor. Got shape: {self.bk.shape}"
            )
        if len(self.ak) != len(self.bk):
            raise ValueError(
                f"ak and bk must have the same length. Got len(ak)={len(self.ak)} and "
                f"len(bk)={len(self.bk)}."
            )
        if self.ak.dtype != self.bk.dtype:
            raise ValueError(
                f"ak and bk must have the same dtype. Got ak.dtype={self.ak.dtype} and "
                f"bk.dtype={self.bk.dtype}."
            )
        if self.ak.device != self.bk.device:
            raise ValueError(
                f"ak and bk must be on the same device. Got ak.device={self.ak.device} "
                f"and bk.device={self.bk.device}."
            )

    def __len__(self):
        """The number of vertical layer interfaces."""
        return len(self.ak)

    def build_corrector(
        self,
        config: AtmosphereCorrectorConfig | CorrectorSelector,
        gridded_operations: GriddedOperations,
        timestep: timedelta,
    ) -> AtmosphereCorrector:
        if (
            isinstance(config, CorrectorSelector)
            and config.type != "atmosphere_corrector"
        ):
            raise ValueError(
                f"Cannot build corrector for vertical coordinate {self} with "
                f"corrector selector {config}."
            )
        if isinstance(config, CorrectorSelector):
            config_instance = dacite.from_dict(
                data_class=AtmosphereCorrectorConfig,
                data=config.config,
                config=dacite.Config(strict=True),
            )
        else:
            config_instance = config
        return AtmosphereCorrector(
            config=config_instance,
            gridded_operations=gridded_operations,
            vertical_coordinate=self,
            timestep=timestep,
        )

    def build_derive_function(self, timestep: timedelta) -> DeriveFnABC:
        return AtmosphericDeriveFn(self, timestep)

    def get_ak(self) -> torch.Tensor:
        return self.ak

    def get_bk(self) -> torch.Tensor:
        return self.bk

    @property
    def coords(self) -> dict[str, np.ndarray]:
        return {"ak": self.ak.cpu().numpy(), "bk": self.bk.cpu().numpy()}

    @property
    def dtype(self) -> torch.dtype:
        return self.ak.dtype

    @property
    def device(self) -> str:
        return self.ak.device

    def to(self, device: str) -> "HybridSigmaPressureCoordinate":
        return HybridSigmaPressureCoordinate(
            ak=self.ak.to(device),
            bk=self.bk.to(device),
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, HybridSigmaPressureCoordinate):
            return False
        try:
            torch.testing.assert_close(self.ak, other.ak)
            torch.testing.assert_close(self.bk, other.bk)
        except AssertionError:
            return False
        return True

    def __repr__(self) -> str:
        return f"HybridSigmaPressureCoordinate(\n    ak={self.ak},\n    bk={self.bk}\n)"

    def as_dict(self) -> TensorMapping:
        return {"ak": self.ak, "bk": self.bk}

    def interface_pressure(self, surface_pressure: torch.Tensor) -> torch.Tensor:
        """
        Compute pressure at vertical layer interfaces.

        Args:
            surface_pressure: The surface pressure in units of Pa.

        Returns:
            A tensor of pressure at vertical layer interfaces. Will contain a new
            dimension at the end, representing the vertical.
        """
        return torch.stack(
            [ak + bk * surface_pressure for ak, bk in zip(self.ak, self.bk)],
            dim=-1,
        )

    def vertical_integral(
        self, integrand: torch.Tensor, surface_pressure: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the mass-weighted vertical integral of the integrand.

        (1 / g) * ∫ x dp

        where
        - g = acceleration due to gravity
        - x = integrad
        - p = pressure level

        Args:
            surface_pressure: The surface pressure in units of Pa.
            integrand: A tensor whose last dimension is the vertical.

        Returns:
            A tensor of same shape as integrand but without the last dimension.
        """
        if len(self.ak) != integrand.shape[-1] + 1:
            raise ValueError(
                "The last dimension of integrand must match the number of vertical "
                "layers in the hybrid sigma-pressure vertical coordinate."
            )
        interface_pressure = self.interface_pressure(surface_pressure)
        pressure_thickness = interface_pressure.diff(dim=-1)
        return (integrand * pressure_thickness).sum(dim=-1) / GRAVITY


LEVEL_PATTERN = re.compile(r"_(\d+)$")


@dataclasses.dataclass
class DepthCoordinate(VerticalCoordinate):
    """
    Defines depth in meters at interface levels and accounts for a constant mask.

    Parameters:
        idepth: depth in meters at interface levels as a 1-dimensional tensor.
        mask: mask indicating valid vertical layers at each spatial point. Must be equal
            to 1 in valid points and 0 elsewhere. The last dimension is the vertical and
            it must be one shorter than idepth. The mask may have additional dimensions
            before the vertical, which are assumed to be broadcastable to match the
            integrand when computing integrals.
    """

    idepth: torch.Tensor
    mask: torch.Tensor
    surface_mask: torch.Tensor | None = None

    def __post_init__(self):
        if len(self.idepth.shape) != 1:
            raise ValueError(
                f"idepth must be a 1-dimensional tensor. Got shape: {self.idepth.shape}"
            )
        if len(self.idepth) < 2:
            raise ValueError(
                f"idepth must have at least two elements. Got {self.idepth}."
            )
        if self.idepth.shape[0] != self.mask.shape[-1] + 1:
            raise ValueError(
                "The last dimension of mask must be one shorter than length of idepth."
                f"Got idepth.shape: {self.idepth.shape} and mask.shape: "
                f"{self.mask.shape}."
            )

    def __len__(self):
        """The number of vertical layer interfaces."""
        return len(self.idepth)

    def get_mask(self) -> torch.Tensor:
        return self.mask

    def get_mask_level(self, level: int) -> torch.Tensor:
        return self.mask.select(dim=-1, index=level)

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        match = LEVEL_PATTERN.search(name)
        if match:
            # 3D variable
            level = int(match.group(1))
            return self.get_mask_level(level)
        else:
            # 2D variable
            if self.surface_mask is not None:
                return self.surface_mask
            return self.get_mask_level(0)

    def get_idepth(self) -> torch.Tensor:
        return self.idepth

    def build_corrector(
        self,
        config: AtmosphereCorrectorConfig | CorrectorSelector,
        gridded_operations: GriddedOperations,
        timestep: timedelta,
    ) -> OceanCorrector:
        if isinstance(config, AtmosphereCorrectorConfig):
            raise ValueError(
                "Cannot build corrector for depth coordinate with an "
                "AtmosphereCorrectorConfig."
            )
        elif config.type != "ocean_corrector":
            raise ValueError(
                f"Cannot build corrector for vertical coordinate {self} with "
                f"corrector selector {config}."
            )
        config_instance = dacite.from_dict(
            data_class=OceanCorrectorConfig,
            data=config.config,
            config=dacite.Config(strict=True),
        )
        return OceanCorrector(
            config=config_instance,
            gridded_operations=gridded_operations,
            vertical_coordinate=self,
            timestep=timestep,
        )

    def build_derive_function(self, timestep: timedelta) -> DeriveFnABC:
        return OceanDeriveFn(self, timestep)

    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]:
        """
        Returns a StaticMasking object that fills in NaNs outside of mask
        valid points, i.e. where the mask value is 0.

        """
        return StaticMasking(
            mask_value=0,
            fill_value=float("nan"),
            mask=self,
        )

    @property
    def dtype(self) -> torch.dtype:
        return self.idepth.dtype

    @property
    def device(self) -> str:
        return self.idepth.device

    @property
    def coords(self) -> dict[str, np.ndarray]:
        return {"idepth": self.idepth.cpu().numpy()}

    def to(self, device: str) -> "DepthCoordinate":
        idepth_on_device = self.idepth.to(device)
        mask_on_device = self.mask.to(device)
        if self.surface_mask is not None:
            surface_mask_on_device = self.surface_mask.to(device)
            return DepthCoordinate(
                idepth_on_device, mask_on_device, surface_mask_on_device
            )
        return DepthCoordinate(idepth_on_device, mask_on_device)

    def __eq__(self, other) -> bool:
        if not isinstance(other, DepthCoordinate):
            return False
        try:
            torch.testing.assert_close(self.idepth, other.idepth)
            torch.testing.assert_close(self.mask, other.mask)
        except AssertionError:
            return False
        return True

    def __repr__(self) -> str:
        return f"DepthCoordinate(\n    idepth={self.idepth},\n    mask={self.mask}\n)"

    def as_dict(self) -> TensorMapping:
        return {"idepth": self.idepth, "mask": self.mask}

    def depth_integral(self, integrand: torch.Tensor) -> torch.Tensor:
        """Compute the depth integral of the integrand.

        ∫ x dz
        where
        - x = integrand
        - z = depth

        Args:
            integrand: A tensor whose last dimension is the vertical.

        Returns:
            A tensor of same shape as integrand but without the last dimension.

        Note:
            NaNs in the integrand are treated as zeros.
        """
        if len(self.idepth) != integrand.shape[-1] + 1:
            raise ValueError(
                f"The last dimension of integrand must match the number of vertical "
                "layers in the depth vertical coordinate. "
                f"Got integrand.shape: {integrand.shape} and idepth.shape: "
                f"{self.idepth.shape}."
            )
        layer_thickness = self.idepth.diff(dim=-1)
        ohc = (integrand * layer_thickness * self.mask).nansum(dim=-1)
        mask = self.get_mask_level(0).expand(ohc.shape)
        return ohc.where(mask > 0, float("nan"))


@dataclasses.dataclass
class NullVerticalCoordinate(VerticalCoordinate):
    """
    A null vertical coordinate system.
    """

    def __eq__(self, other) -> bool:
        return isinstance(other, NullVerticalCoordinate)

    def __repr__(self) -> str:
        return "NullVerticalCoordinate()"

    def __len__(self) -> int:
        return 0

    def build_corrector(
        self,
        config: AtmosphereCorrectorConfig | CorrectorSelector,
        gridded_operations: GriddedOperations,
        timestep: timedelta,
    ) -> CorrectorABC:
        if isinstance(config, AtmosphereCorrectorConfig):
            return AtmosphereCorrector(
                config=config,
                gridded_operations=gridded_operations,
                vertical_coordinate=None,
                timestep=timestep,
            )
        if config.type == "atmosphere_corrector":
            config_instance = dacite.from_dict(
                data_class=AtmosphereCorrectorConfig,
                data=config.config,
                config=dacite.Config(strict=True),
            )
            return AtmosphereCorrector(
                config=config_instance,
                gridded_operations=gridded_operations,
                vertical_coordinate=None,
                timestep=timestep,
            )
        elif config.type == "ocean_corrector":
            config_instance = dacite.from_dict(
                data_class=OceanCorrectorConfig,
                data=config.config,
                config=dacite.Config(strict=True),
            )
            return OceanCorrector(
                config=config_instance,
                gridded_operations=gridded_operations,
                vertical_coordinate=None,
                timestep=timestep,
            )
        else:
            raise ValueError(
                f"Invalid corrector type: {config.type}. "
                "Must be either 'atmosphere_corrector' or 'ocean_corrector'."
            )

    def build_derive_function(self, timestep: timedelta) -> DeriveFnABC:
        return NullDeriveFn()

    def to(self, device: str) -> "NullVerticalCoordinate":
        return self

    def as_dict(self) -> TensorMapping:
        return {}

    @property
    def coords(self) -> dict[str, np.ndarray]:
        return {}


OptionalHybridSigmaPressureCoordinate = (
    HybridSigmaPressureCoordinate | NullVerticalCoordinate
)

OptionalDepthCoordinate = DepthCoordinate | NullVerticalCoordinate


@dataclasses.dataclass
class SerializableVerticalCoordinate:
    """Only for use in serializing/deserializing coordinates with dacite."""

    vertical_coordinate: (
        HybridSigmaPressureCoordinate | DepthCoordinate | NullVerticalCoordinate
    )

    @classmethod
    def from_state(cls, state) -> VerticalCoordinate:
        return dacite.from_dict(
            data_class=cls,
            data={"vertical_coordinate": state},
            config=dacite.Config(strict=True),
        ).vertical_coordinate


@dataclasses.dataclass
class DimSize:
    name: str
    size: int


class HorizontalCoordinates(abc.ABC):
    """
    Parent class for horizontal coordinate system grids.
    Contains coords which must be subclassed to provide the coordinates.
    """

    SelfType = TypeVar("SelfType", bound="HorizontalCoordinates")

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    @abc.abstractmethod
    def to(self: HC, device: str) -> HC:
        pass

    @property
    @abc.abstractmethod
    def coords(self) -> Mapping[str, np.ndarray]:
        pass

    @property
    @abc.abstractmethod
    def xyz(self) -> tuple[float, float, float]:
        pass

    @property
    @abc.abstractmethod
    def dims(self) -> list[str]:
        """Names of model horizontal dimensions."""
        pass

    @property
    @abc.abstractmethod
    def loaded_sizes(self) -> list[DimSize]:
        """Sizes of horizontal dimensions as loaded from training dataset."""
        pass

    @property
    @abc.abstractmethod
    def grid(self) -> Literal["equiangular", "legendre-gauss", "healpix"]:
        pass

    @property
    @abc.abstractmethod
    def area_weights(self) -> torch.Tensor | None:
        pass

    @abc.abstractmethod
    def get_gridded_operations(
        self, mask_provider: MaskProviderABC = NullMaskProvider
    ) -> GriddedOperations:
        pass

    @property
    @abc.abstractmethod
    def meshgrid(self) -> tuple[torch.Tensor, torch.Tensor]:
        """Meshgrids of latitudes and longitudes, respectively."""
        pass

    @property
    @abc.abstractmethod
    def shape(self) -> tuple[int, ...]:
        pass

    @abc.abstractmethod
    def to_state(self) -> TensorMapping:
        pass


@dataclasses.dataclass
class LatLonCoordinates(HorizontalCoordinates):
    """
    Defines a (latitude, longitude) grid.

    Parameters:
        lat: 1-dimensional tensor of latitudes
        lon: 1-dimensional tensor of longitudes
    """

    lon: torch.Tensor
    lat: torch.Tensor

    def __post_init__(self):
        self._area_weights: torch.Tensor | None = None

    def __eq__(self, other) -> bool:
        if not isinstance(other, LatLonCoordinates):
            return False
        lat_eq = torch.allclose(self.lat, other.lat)
        lon_eq = torch.allclose(self.lon, other.lon)
        return lat_eq and lon_eq

    def __repr__(self) -> str:
        return f"LatLonCoordinates(\n    lat={self.lat},\n    lon={self.lon}\n"

    def to(self, device: str) -> "LatLonCoordinates":
        return LatLonCoordinates(
            lon=self.lon.to(device),
            lat=self.lat.to(device),
        )

    @property
    def area_weights(self) -> torch.Tensor:
        if self._area_weights is None:
            self._area_weights = metrics.spherical_area_weights(self.lat, len(self.lon))
        return self._area_weights

    @property
    def coords(self) -> Mapping[str, np.ndarray]:
        return {
            self.dims[0]: self.lat.cpu().type(torch.float32).numpy(),
            self.dims[1]: self.lon.cpu().type(torch.float32).numpy(),
        }

    @property
    def xyz(self) -> tuple[float, float, float]:
        lats, lons = np.broadcast_arrays(
            self.coords[self.dims[0]][:, None], self.coords[self.dims[1]][None, :]
        )
        return lon_lat_to_xyz(lons, lats)

    @property
    def dims(self) -> list[str]:
        return ["lat", "lon"]

    @property
    def loaded_sizes(self) -> list[DimSize]:
        return [
            DimSize(self.dims[0], len(self.lat)),
            DimSize(self.dims[1], len(self.lon)),
        ]

    @property
    def grid(self) -> Literal["equiangular", "legendre-gauss"]:
        if torch.allclose(
            self.lat[1:] - self.lat[:-1],
            self.lat[1] - self.lat[0],
        ):
            return "equiangular"
        else:
            return "legendre-gauss"

    def get_gridded_operations(
        self, mask_provider: MaskProviderABC = NullMaskProvider
    ) -> LatLonOperations:
        return LatLonOperations(self.area_weights, mask_provider)

    @property
    def meshgrid(self) -> tuple[torch.Tensor, torch.Tensor]:
        return torch.meshgrid(self.lat, self.lon, indexing="ij")

    @property
    def shape(self) -> tuple[int, int]:
        return (len(self.lat), len(self.lon))

    def to_state(self) -> TensorMapping:
        return {"lat": self.lat, "lon": self.lon}


@dataclasses.dataclass
class HEALPixCoordinates(HorizontalCoordinates):
    """
    Defines a HEALPix (face, height, width) grid. See https://healpix.jpl.nasa.gov/ for
    more information.

    Parameters:
        face: 1-dimensional tensor of faces
        height: 1-dimensional tensor of heights
        width: 1-dimensional tensor of widths
    """

    face: torch.Tensor
    height: torch.Tensor
    width: torch.Tensor

    def __eq__(self, other) -> bool:
        if not isinstance(other, HEALPixCoordinates):
            return False
        return (
            torch.allclose(self.face, other.face)
            and torch.allclose(self.height, other.height)
            and torch.allclose(self.width, other.width)
        )

    def __repr__(self) -> str:
        return (
            "HEALPixCoordinates(\n"
            f"    face={self.face},\n"
            f"    height={self.height},\n"
            f"    width={self.width}\n"
            ")"
        )

    def to(self, device: str) -> "HEALPixCoordinates":
        return HEALPixCoordinates(
            face=self.face.to(device),
            height=self.height.to(device),
            width=self.width.to(device),
        )

    @property
    def coords(self) -> Mapping[str, np.ndarray]:
        return {
            "face": self.face.cpu().type(torch.float32).numpy(),
            "height": self.height.cpu().type(torch.float32).numpy(),
            "width": self.width.cpu().type(torch.float32).numpy(),
        }

    @property
    def xyz(self) -> tuple[float, float, float]:
        level = int(math.log2(len(self.width)))
        hpx = e2ghpx.Grid(level=level, pixel_order=e2ghpx.HEALPIX_PAD_XY)
        lats = hpx.lat
        lats = lats.reshape(len(self.face), len(self.width), len(self.height))
        lons = hpx.lon
        lons = lons.reshape(len(self.face), len(self.width), len(self.height))
        x, y, z = lon_lat_to_xyz(lat=lats, lon=lons)
        return x, y, z

    @property
    def dims(self) -> list[str]:
        return ["face", "height", "width"]

    @property
    def loaded_dims(self) -> list[str]:
        return self.dims

    @property
    def loaded_sizes(self) -> list[DimSize]:
        return [
            DimSize("face", len(self.face)),
            DimSize("height", len(self.width)),
            DimSize("width", len(self.height)),
        ]

    @property
    def grid(self) -> Literal["healpix"]:
        return "healpix"

    @property
    def area_weights(self) -> Literal[None]:
        return None

    def get_gridded_operations(
        self, mask_provider: MaskProviderABC = NullMaskProvider
    ) -> HEALPixOperations:
        # this code is necessary because when no masks are in a given dataset, we return
        # an empty MaskProvider instead of the NullMaskProvider.
        if mask_provider == NullMaskProvider:
            null_mask = True
        elif isinstance(mask_provider, MaskProvider):
            null_mask = len(mask_provider.masks) == 0
        else:
            raise TypeError(
                f"Don't know how to handle given mask_provider: {mask_provider}"
            )
        if not (null_mask):
            raise NotImplementedError(
                "HEALPixCoordinates does not support a mask provider. "
                "Use NullMaskProvider when getting gridded operations "
                "for HEALPixCoordinates."
            )
        return HEALPixOperations()

    @property
    def meshgrid(self) -> tuple[torch.Tensor, torch.Tensor]:
        # We'll return a 3D (face, width, height) tensor representing the lat-lon
        # coordinates of this grid.
        level = int(math.log2(len(self.width)))
        hpx = e2ghpx.Grid(level=level, pixel_order=e2ghpx.HEALPIX_PAD_XY)
        lats = hpx.lat
        lats = lats.reshape(self.shape)
        lons = hpx.lon
        lons = lons.reshape(self.shape)
        return lats, lons

    @property
    def shape(self) -> tuple[int, int, int]:
        return (len(self.face), len(self.width), len(self.height))

    def to_state(self) -> TensorMapping:
        return {"face": self.face, "height": self.height, "width": self.width}


@dataclasses.dataclass
class SerializableHorizontalCoordinates:
    """Only for use in serializing/deserializing coordinates with dacite."""

    horizontal_coordinates: LatLonCoordinates | HEALPixCoordinates

    @classmethod
    def from_state(cls, state) -> HorizontalCoordinates:
        return dacite.from_dict(
            data_class=cls,
            data={"horizontal_coordinates": state},
            config=dacite.Config(strict=True),
        ).horizontal_coordinates



import pytest
import torch

from fme import get_device
from fme.ace.aggregator.one_step.reduced_metrics import AreaWeightedReducedMetric
from fme.core.typing_ import TensorDict, TensorMapping

DEVICE = get_device()


def mock_area_weighted_metric(
    truth: TensorMapping, predicted: TensorMapping
) -> TensorDict:
    result = {}
    for name in truth:
        diff = predicted[name] - truth[name]
        result[name] = diff.mean(dim=list(range(1, diff.ndim)))
    return result


def test_area_weighted_reduced_metric_record_and_get():
    metric = AreaWeightedReducedMetric(
        device=DEVICE, compute_metric=mock_area_weighted_metric
    )

    # Batch 1
    target1 = {
        "a": torch.ones(2, 3, 4, 5, device=DEVICE),
        "b": torch.zeros(2, 3, 4, 5, device=DEVICE),
    }
    gen1 = {
        "a": torch.full((2, 3, 4, 5), 2.0, device=DEVICE),
        "b": torch.ones(2, 3, 4, 5, device=DEVICE),
    }
    # Expected batch 1 means: a=(2-1)=1, b=(1-0)=1
    metric.record(target1, gen1)

    # Batch 2
    target2 = {
        "a": torch.ones(2, 3, 4, 5, device=DEVICE),
        "b": torch.zeros(2, 3, 4, 5, device=DEVICE),
    }
    gen2 = {
        "a": torch.full((2, 3, 4, 5), 4.0, device=DEVICE),
        "b": torch.full((2, 3, 4, 5), 2.0, device=DEVICE),
    }
    # Expected batch 2 means: a=(4-1)=3, b=(2-0)=2
    metric.record(target2, gen2)

    # AreaWeightedReducedMetric sums the batch averages
    # Expected total: a = 1 + 3, b = 1 + 2
    expected_total = {
        "a": torch.tensor(1.0 + 3.0, device=DEVICE),
        "b": torch.tensor(1.0 + 2.0, device=DEVICE),
    }

    result = metric.get()

    assert result.keys() == expected_total.keys()
    torch.testing.assert_close(result["a"], expected_total["a"])
    torch.testing.assert_close(result["b"], expected_total["b"])


def test_area_weighted_reduced_metric_new_keys_raises():
    metric = AreaWeightedReducedMetric(
        device=DEVICE, compute_metric=mock_area_weighted_metric
    )

    target1 = {"a": torch.ones(1, 1, device=DEVICE)}
    gen1 = {"a": torch.ones(1, 1, device=DEVICE)}
    metric.record(target1, gen1)

    target2 = {
        "a": torch.ones(1, 1, device=DEVICE),
        "b": torch.ones(1, 1, device=DEVICE),
    }
    gen2 = {"a": torch.ones(1, 1, device=DEVICE), "b": torch.ones(1, 1, device=DEVICE)}
    with pytest.raises(ValueError, match="'b'"):
        metric.record(target2, gen2)


def test_area_weighted_reduced_metric_missing_keys_raises():
    metric = AreaWeightedReducedMetric(
        device=DEVICE, compute_metric=mock_area_weighted_metric
    )

    target1 = {
        "a": torch.ones(1, 1, device=DEVICE),
        "b": torch.ones(1, 1, device=DEVICE),
    }
    gen1 = {"a": torch.ones(1, 1, device=DEVICE), "b": torch.ones(1, 1, device=DEVICE)}
    metric.record(target1, gen1)

    target2 = {"a": torch.ones(1, 1, device=DEVICE)}
    gen2 = {"a": torch.ones(1, 1, device=DEVICE)}
    with pytest.raises(ValueError, match="'b'"):
        metric.record(target2, gen2)


def test_area_weighted_reduced_metric_empty():
    metric = AreaWeightedReducedMetric(
        device=DEVICE, compute_metric=mock_area_weighted_metric
    )
    metrics = metric.get()
    assert torch.isnan(metrics["anything"])



import dataclasses
import unittest.mock

import pytest
import torch

from fme.core.dataset_info import DatasetInfo
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import StandardNormalizer

from .multi_call import (
    MultiCallStepConfig,
    _extend_normalizer_with_multi_call_outputs,
    replace_multi_call,
)
from .step import StepSelector
from .test_step_registry import MockStep, MockStepConfig


@pytest.mark.parametrize("include_multi_call_in_loss", [True, False])
def test_multi_call(include_multi_call_in_loss: bool):
    output_names = ["b", "c"]
    multi_call_output_names = ["c"]

    def _step(input, next_step_input_data, wrapper):
        prediction = {k: input["CO2"].detach().clone() for k in output_names}
        return prediction

    config = MultiCallStepConfig(
        wrapped_step=StepSelector(
            type="mock",
            config={"in_names": ["CO2", "a", "b"], "out_names": ["b", "c"]},
        ),
        config=MultiCallConfig(
            forcing_name="CO2",
            forcing_multipliers={"_doubled_co2": 2},
            output_names=multi_call_output_names,
        ),
        include_multi_call_in_loss=include_multi_call_in_loss,
    )

    with unittest.mock.patch.object(MockStep, "step", side_effect=_step):
        step = config.get_step(DatasetInfo(), lambda x: None)
        assert step.output_names == ["b", "c", "c_doubled_co2"]
        if include_multi_call_in_loss:
            assert step.loss_names == ["b", "c", "c_doubled_co2"]
        else:
            assert step.loss_names == ["b", "c"]

        input = {
            "a": torch.randn(1, 2, 3, 4),
            "b": torch.randn(1, 2, 3, 4),
            "CO2": torch.randn(1, 2, 3, 4),
        }
        out = step.step(input, {})
    torch.testing.assert_close(out["b"], input["CO2"])
    torch.testing.assert_close(out["c"], input["CO2"])
    torch.testing.assert_close(out["c_doubled_co2"], input["CO2"] * 2)


def test_replace_multi_call_adds_wrapper():
    selector = StepSelector(
        type="mock", config={"in_names": ["a", "b"], "out_names": ["c", "d"]}
    )
    multi_call = MultiCallConfig(
        forcing_name="a",
        forcing_multipliers={"_doubled_a": 2},
        output_names=["c"],
    )
    state = {"foo": "bar"}
    new_selector, new_state = replace_multi_call(selector, multi_call, state)
    assert new_selector.type == "multi_call"
    assert new_selector.config["wrapped_step"] == dataclasses.asdict(selector)
    assert new_selector.config["config"] == dataclasses.asdict(multi_call)
    new_step_config = new_selector._step_config_instance
    assert isinstance(new_step_config, MultiCallStepConfig)
    assert new_step_config.wrapped_step == selector
    assert new_step_config.config == multi_call
    assert new_step_config.include_multi_call_in_loss
    assert new_state["wrapped_step"] == state


def test_replace_multi_call_updates_wrapper():
    mock_selector = StepSelector(
        type="mock", config={"in_names": ["a", "b"], "out_names": ["c", "d"]}
    )
    multi_call = MultiCallConfig(
        forcing_name="a",
        forcing_multipliers={"_doubled_a": 2},
        output_names=["c"],
    )
    selector = StepSelector(
        type="multi_call",
        config={
            "wrapped_step": dataclasses.asdict(mock_selector),
            "config": dataclasses.asdict(multi_call),
        },
    )
    new_multi_call = MultiCallConfig(
        forcing_name="b",
        forcing_multipliers={"_doubled_b": 2},
        output_names=["d"],
    )
    state = {"wrapped_step": {"foo": "bar"}}
    new_selector, new_state = replace_multi_call(selector, new_multi_call, state)
    assert new_selector.type == "multi_call"
    assert new_selector.config["wrapped_step"] == dataclasses.asdict(mock_selector)
    assert new_selector.config["config"] == dataclasses.asdict(new_multi_call)
    new_step_config = new_selector._step_config_instance
    assert isinstance(new_step_config, MultiCallStepConfig)
    assert isinstance(new_step_config.wrapped_step, StepSelector)
    assert new_step_config.wrapped_step.type == "mock"
    assert new_step_config.wrapped_step == mock_selector
    assert new_step_config.config == new_multi_call
    assert new_state == state


def test_replace_multi_call_updates_wrapper_with_none():
    mock_selector = StepSelector(
        type="mock", config={"in_names": ["a", "b"], "out_names": ["c", "d"]}
    )
    multi_call = MultiCallConfig(
        forcing_name="a",
        forcing_multipliers={"_doubled_a": 2},
        output_names=["c"],
    )
    selector = StepSelector(
        type="multi_call",
        config={
            "wrapped_step": dataclasses.asdict(mock_selector),
            "config": dataclasses.asdict(multi_call),
        },
    )
    state = {"wrapped_step": {"foo": "bar"}}
    new_selector, new_state = replace_multi_call(selector, None, state)
    assert new_selector.type == "multi_call"
    assert new_selector.config["wrapped_step"] == dataclasses.asdict(mock_selector)
    assert new_selector.config["config"] is None
    new_step_config = new_selector._step_config_instance
    assert isinstance(new_step_config, MultiCallStepConfig)
    assert new_step_config.wrapped_step.type == "mock"
    assert new_step_config.wrapped_step == mock_selector
    assert new_step_config.config is None
    assert new_state == state


def test_extend_normalizer_with_multi_call_outputs():
    torch.manual_seed(0)
    c_mean, c_std = torch.randn(2)
    normalizer = StandardNormalizer(
        means={"a": torch.tensor(0), "b": torch.tensor(0), "c": c_mean},
        stds={"a": torch.tensor(1), "b": torch.tensor(1), "c": c_std},
    )
    config = MultiCallConfig(
        forcing_name="CO2",
        forcing_multipliers={"_doubled_co2": 2, "_halved_co2": 0.5},
        output_names=["c"],
    )
    result = _extend_normalizer_with_multi_call_outputs(config, normalizer)
    assert result.means == {
        "a": torch.tensor(0),
        "b": torch.tensor(0),
        "c": c_mean,
        "c_doubled_co2": c_mean,
        "c_halved_co2": c_mean,
    }
    assert result.stds == {
        "a": torch.tensor(1),
        "b": torch.tensor(1),
        "c": c_std,
        "c_doubled_co2": c_std,
        "c_halved_co2": c_std,
    }


@pytest.mark.parametrize("include_multi_call_in_loss", [True, False])
def test_loss_normalizer_uses_extra_stats_names(include_multi_call_in_loss: bool):
    torch.manual_seed(0)
    a_mean, a_std = torch.randn(2)
    b_mean, b_std = torch.randn(2)
    b_doubled_mean, b_doubled_std = torch.randn(2)
    b_halved_mean, b_halved_std = torch.randn(2)
    c_mean, c_std = torch.randn(2)
    d_mean, d_std = torch.randn(2)
    c_doubled_mean, c_doubled_std = torch.randn(2)
    c_halved_mean, c_halved_std = torch.randn(2)
    means = {
        "a": a_mean,
        "b": b_mean,
        "b_doubled_co2": b_doubled_mean,
        "b_halved_co2": b_halved_mean,
        "c": c_mean,
        "c_doubled_co2": c_doubled_mean,
        "c_halved_co2": c_halved_mean,
        "d": d_mean,
    }
    stds = {
        "a": a_std,
        "b": b_std,
        "b_doubled_co2": b_doubled_std,
        "b_halved_co2": b_halved_std,
        "c": c_std,
        "c_doubled_co2": c_doubled_std,
        "c_halved_co2": c_halved_std,
        "d": d_std,
    }

    def _get_loss_normalizer(
        extra_names: list[str] | None,
        extra_residual_scaled_names: list[str] | None,
    ):
        base_names = ["a", "b", "c", "d"]
        if extra_names is None:
            extra_names = []
        if extra_residual_scaled_names is None:
            extra_residual_scaled_names = []
        extra_names = extra_names + extra_residual_scaled_names
        return StandardNormalizer(
            means={
                **{k: means[k] for k in base_names},
                **{k: means[k] for k in extra_names},
            },
            stds={
                **{k: stds[k] for k in base_names},
                **{k: stds[k] for k in extra_names},
            },
        )

    mock_get_loss_normalizer = unittest.mock.Mock(side_effect=_get_loss_normalizer)
    with unittest.mock.patch.object(
        MockStepConfig, "get_loss_normalizer", side_effect=mock_get_loss_normalizer
    ):
        config = MultiCallStepConfig(
            wrapped_step=StepSelector(
                type="mock",
                config={"in_names": ["CO2", "a", "b"], "out_names": ["b", "c", "d"]},
            ),
            config=MultiCallConfig(
                forcing_name="CO2",
                forcing_multipliers={"_doubled_co2": 2, "_halved_co2": 0.5},
                output_names=["b", "c"],
            ),
            # we expect multi-call in loss normalizer regardless of
            # include_multi_call_in_loss
            include_multi_call_in_loss=include_multi_call_in_loss,
        )
        normalizer = config.get_loss_normalizer()
        assert normalizer.means == {
            "a": a_mean,
            "b": b_mean,
            "b_doubled_co2": b_doubled_mean,
            "b_halved_co2": b_halved_mean,
            "c": c_mean,
            "c_doubled_co2": c_doubled_mean,
            "c_halved_co2": c_halved_mean,
            "d": d_mean,
        }
        assert normalizer.stds == {
            "a": a_std,
            "b": b_std,
            "b_doubled_co2": b_doubled_std,
            "b_halved_co2": b_halved_std,
            "c": c_std,
            "c_doubled_co2": c_doubled_std,
            "c_halved_co2": c_halved_std,
            "d": d_std,
        }
        assert mock_get_loss_normalizer.call_count == 1
        call_args = mock_get_loss_normalizer.call_args
        assert set(call_args.kwargs["extra_names"]) == {
            "b_doubled_co2",
            "b_halved_co2",
            "c_doubled_co2",
            "c_halved_co2",
        }
        assert set(call_args.kwargs["extra_residual_scaled_names"]) == {
            "b_doubled_co2",
            "b_halved_co2",
        }



from datetime import timedelta

import cftime
import pytest
import torch
import xarray as xr
from matplotlib import pyplot as plt

from fme import get_device
from fme.core.coordinates import LatLonCoordinates
from fme.core.testing import mock_distributed
from fme.core.typing_ import TensorMapping

from .dynamic_index import (
    LatLonRegion,
    PairedRegionalIndexAggregator,
    RegionalIndexAggregator,
    anomalies_from_monthly_climo,
    running_monthly_mean,
)


def _get_windowed_data(
    n_samples: int,
    n_times: int,
    n_lat: int,
    n_lon: int,
    i_start: int = 0,
    sst_name: str = "surface_temperature",
    time_varying: bool = True,
) -> TensorMapping:
    lat_coord = torch.linspace(0.0, 10.0, n_lat)
    lon_coord = torch.linspace(0.0, 20.0, n_lon)
    sst_data = (
        (lat_coord.unsqueeze(-1) + lon_coord.unsqueeze(0))
        .unsqueeze(0)
        .unsqueeze(0)
        .expand(n_samples, n_times, n_lat, n_lon)
    )
    if time_varying:
        sst_data = (
            sst_data
            + torch.arange(i_start, i_start + n_times, 1.0)[None, :, None, None]
        )
    return {
        sst_name: sst_data.to(device=get_device()),
        "lat": lat_coord,
        "lon": lon_coord,
    }


def _get_windowed_times(
    start_time: tuple[int, ...],
    n_samples: int,
    n_times: int,
    i_start: int = 0,
    freq="6h",
    calendar: str = "noleap",
) -> xr.DataArray:
    start_time = cftime.datetime(*start_time, calendar=calendar) + timedelta(
        hours=6 * i_start
    )
    sample_time_array = xr.DataArray(
        data=xr.date_range(
            start=start_time,
            periods=n_times,
            freq=freq,
            use_cftime=True,
        ).values,
        dims=("time",),
    )
    return xr.concat(
        [sample_time_array for i in range(n_samples)],
        dim="sample",
    )


@pytest.mark.parametrize(
    "lat_bounds, lon_bounds, case",
    [
        pytest.param((0.0, 10.0), (0.0, 20.0), "original_domain"),
        pytest.param((20.0, 30.0), (25.0, 35.0), "null_domain"),
        pytest.param((0.0, 5.0), (0.0, 20.0), "first_half_lat"),
        pytest.param((0.0, 10.0), (0.0, 10.0), "first_half_lon"),
        pytest.param((0.0, 5.0), (0.0, 10.0), "first_half_both"),
    ],
)
def test_lat_lon_region(lat_bounds, lon_bounds, case):
    n_lat, n_lon = 3, 5
    lat_coord = torch.linspace(0.0, 10.0, n_lat)
    lon_coord = torch.linspace(0.0, 20.0, n_lon)
    region = LatLonRegion(
        lat=lat_coord,
        lon=lon_coord,
        lat_bounds=lat_bounds,
        lon_bounds=lon_bounds,
    )
    regional_weights = region.regional_weights
    assert regional_weights.shape == (n_lat, n_lon)
    if case == "original_domain":
        assert torch.allclose(regional_weights, torch.ones_like(regional_weights))
    elif case == "null_domain":
        assert torch.allclose(regional_weights, torch.zeros_like(regional_weights))
    elif case == "first_half_lat":
        expected = torch.tensor(
            [
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, 0],
            ],
            dtype=torch.float32,
        )
        assert torch.allclose(regional_weights, expected)
    elif case == "first_half_lon":
        expected = torch.tensor(
            [
                [1, 1, 1, 0, 0],
                [1, 1, 1, 0, 0],
                [1, 1, 1, 0, 0],
            ],
            dtype=torch.float32,
        )
        assert torch.allclose(regional_weights, expected)
    else:
        expected = torch.tensor(
            [
                [1, 1, 1, 0, 0],
                [1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0],
            ],
            dtype=torch.float32,
        )
        assert torch.allclose(regional_weights, expected)


def test_regional__raw_index():
    n_samples, n_times, n_lat, n_lon = 2, 10, 11, 21
    start_date = (2000, 1, 1, 0, 0, 0)
    i_start = 0
    first_data = _get_windowed_data(
        n_samples,
        n_times // 2,
        n_lat,
        n_lon,
        i_start,
    )
    lat_lon_coordinates = LatLonCoordinates(
        lat=first_data["lat"],
        lon=first_data["lon"],
    )
    region = LatLonRegion(
        lat=lat_lon_coordinates.lat,
        lon=lat_lon_coordinates.lon,
        lat_bounds=(4.5, 6.5),
        lon_bounds=(9.5, 11.5),
    )
    # overwrite the area weights with ones for testing
    gridded_operations = lat_lon_coordinates.get_gridded_operations()
    gridded_operations._cpu_area = torch.ones(n_lat, n_lon)
    gridded_operations._device_area = torch.ones(n_lat, n_lon).to(device=get_device())
    agg = RegionalIndexAggregator(
        regional_weights=region.regional_weights,
        regional_mean=gridded_operations.regional_area_weighted_mean,
    )
    first_times = _get_windowed_times(start_date, n_samples, n_times // 2, i_start)
    agg.record_batch(first_times, first_data)
    i_start += n_times // 2
    second_data = _get_windowed_data(
        n_samples,
        n_times // 2,
        n_lat,
        n_lon,
        i_start,
    )
    second_times = _get_windowed_times(start_date, n_samples, n_times // 2, i_start)
    agg.record_batch(second_times, second_data)

    expected_values = torch.arange(16.0, 16.0 + n_times, 1.0).to(device=get_device())
    raw_indices: torch.Tensor = agg._raw_indices
    for raw_index in raw_indices.values():
        assert raw_index.shape == (n_samples, n_times)
        assert torch.allclose(raw_index, expected_values)
    raw_times: xr.DataArray = agg._raw_index_times
    assert raw_times.sizes["sample"] == n_samples
    assert raw_times.sizes["time"] == n_times
    expected_times = xr.concat(
        [
            xr.DataArray(
                data=xr.date_range(
                    start=cftime.DatetimeNoLeap(*start_date),
                    periods=n_times,
                    freq="6h",
                    use_cftime=True,
                ).values,
                dims=("time"),
            )
            for _ in range(n_samples)
        ],
        dim="sample",
    )
    xr.testing.assert_allclose(raw_times, expected_times)


def test_anomalies_from_monthly_climo():
    n_months = 24
    n_samples = 2
    # data starts at 0, increases by 1.0 every month
    data = _get_windowed_data(
        n_samples=n_samples,
        n_times=n_months,
        n_lat=1,
        n_lon=1,
    )["surface_temperature"].squeeze()
    times = _get_windowed_times(
        start_time=(2000, 1, 1, 0, 0, 0), n_samples=2, n_times=n_months, freq="MS"
    )
    anomalies = anomalies_from_monthly_climo(data, times)
    assert anomalies.shape == data.shape
    assert torch.allclose(
        anomalies.mean(dim=1), torch.zeros(n_samples).to(device=get_device())
    )
    expected = torch.concat(
        [
            torch.full((n_samples, 12), fill_value=-6.0),
            torch.full((n_samples, 12), fill_value=6.0),
        ],
        dim=1,
    ).to(device=get_device())
    assert torch.allclose(anomalies, expected)


def test_running_monthly_mean_monthly_averaging():
    n_times = 1440  # 12 months of 6 hourly, 30-day months
    n_samples = 2
    # data starts at 0, increases by 1.0 every 6 hours
    data = _get_windowed_data(
        n_samples=n_samples,
        n_times=n_times,
        n_lat=1,
        n_lon=1,
    )["surface_temperature"].squeeze()
    times = _get_windowed_times(
        start_time=(2000, 1, 1, 0, 0, 0),
        n_samples=2,
        n_times=n_times,
        freq="6h",
        calendar="360_day",
    )
    running_mean, _ = running_monthly_mean(data, times, n_months=1)
    assert running_mean.shape == (n_samples, 12)
    expected = (  # monthly means for 30-day months
        torch.arange(0.0, n_times, 1.0)
        .reshape(12, 120)
        .mean(dim=1)
        .unsqueeze(0)
        .expand(n_samples, -1)
    ).to(device=get_device())
    assert torch.allclose(running_mean, expected)


@pytest.mark.parametrize(
    ["n_months"], [pytest.param(1, id="1-month"), pytest.param(5, id="5-month")]
)
def test_running_monthly_mean_window_averaging(n_months: int):
    n_times = 24
    n_samples = 2
    # data starts at 0, increases by 1.0 every month
    data = _get_windowed_data(
        n_samples=n_samples,
        n_times=n_times,
        n_lat=1,
        n_lon=1,
    )["surface_temperature"].squeeze()
    times = _get_windowed_times(
        start_time=(2000, 1, 1, 0, 0, 0), n_samples=2, n_times=n_times, freq="MS"
    )
    running_mean, _ = running_monthly_mean(data, times, n_months=n_months)
    assert running_mean.shape == data.shape
    expected = torch.cat(
        [  # first n_months - 1 months are nan
            torch.full((n_samples, n_months - 1), fill_value=float("nan")),
            # remaining months are the running mean of n_months months
            torch.arange(n_months / 2.0 - 0.5, n_times - n_months / 2.0 + 0.5, 1.0)
            .unsqueeze(0)
            .expand(n_samples, -1),
        ],
        dim=1,
    ).to(device=get_device())
    assert torch.allclose(running_mean, expected, equal_nan=True)


@pytest.mark.parametrize(
    "use_mock_distributed",
    [pytest.param(False, id="single_process"), pytest.param(True, id="distributed")],
)
def test_regional_index__get_gathered_indices(use_mock_distributed):
    n_samples, n_times, n_lat, n_lon = 1, 1440, 11, 21
    start_date = (2000, 1, 1, 0, 0, 0)
    first_sample_data = _get_windowed_data(n_samples, n_times, n_lat, n_lon)
    second_sample_data = _get_windowed_data(
        n_samples, n_times, n_lat, n_lon, i_start=n_times
    )
    sample_data = {
        "surface_temperature": torch.concat(
            [
                first_sample_data["surface_temperature"],
                second_sample_data["surface_temperature"],
            ],
            dim=0,
        ),
    }
    first_sample_times = _get_windowed_times(
        start_date, n_samples, n_times, i_start=0
    )  # first sample's inference period spans year 2000
    second_sample_times = _get_windowed_times(
        start_date, n_samples, n_times, i_start=n_times
    )  # second sample's inference period spans year 2001
    sample_times = xr.concat([first_sample_times, second_sample_times], dim="sample")
    lat_lon_coordinates = LatLonCoordinates(
        lat=first_sample_data["lat"],
        lon=first_sample_data["lon"],
    )
    region = LatLonRegion(
        lat=lat_lon_coordinates.lat,
        lon=lat_lon_coordinates.lon,
        lat_bounds=(4.5, 6.5),
        lon_bounds=(9.5, 11.5),
    )
    agg = RegionalIndexAggregator(
        regional_weights=region.regional_weights,
        regional_mean=lat_lon_coordinates.get_gridded_operations().regional_area_weighted_mean,
    )
    agg.record_batch(sample_times, sample_data)
    if use_mock_distributed:
        world_size = 2
        with mock_distributed(world_size=world_size):
            indices_dataset = agg.get_dataset()
            assert indices_dataset is not None
    else:
        world_size = 1
        indices_dataset = agg.get_dataset()
        assert indices_dataset is not None
    assert set(indices_dataset.dims) == {"sample", "time"}
    assert set(indices_dataset.time.dt.year.values) == {2000, 2001}
    assert indices_dataset.sizes["sample"] == 2 * n_samples * world_size
    assert indices_dataset.sizes["time"] == 2 * 12  # 2 years of 12 months each


@pytest.mark.parametrize(
    "variable_name",
    [
        pytest.param("surface_temperature", id="surface_temperature"),
        pytest.param("sst", id="sst"),
        pytest.param("TS", id="TS"),
    ],
)
def test_regional_index_aggregator(variable_name):
    n_lat = 10
    n_lon = 20
    n_sample = 2
    n_times = 365 * 4  # one year of data for monthly averaging
    data = _get_windowed_data(n_sample, n_times, n_lat, n_lon, sst_name=variable_name)
    time = _get_windowed_times((2000, 1, 1, 0, 0, 0), n_sample, n_times)
    lat_lon_coordinates = LatLonCoordinates(
        lat=data["lat"],
        lon=data["lon"],
    )
    region = LatLonRegion(
        lat=lat_lon_coordinates.lat,
        lon=lat_lon_coordinates.lon,
        lat_bounds=(4.5, 6.5),
        lon_bounds=(9.5, 11.5),
    )
    agg = RegionalIndexAggregator(
        regional_weights=region.regional_weights,
        regional_mean=lat_lon_coordinates.get_gridded_operations().regional_area_weighted_mean,
    )
    agg.record_batch(time=time, data=data)
    logs = agg.get_logs(label="test")
    assert len(logs) > 0
    metric_name = f"test/{variable_name}_nino34_index"
    assert metric_name in logs
    assert isinstance(logs[metric_name], plt.Figure)

    metric_name = f"test/{variable_name}_nino34_index_power_spectrum"
    assert metric_name in logs
    assert isinstance(logs[metric_name], plt.Figure)


@pytest.mark.parametrize(
    "variable_name",
    [
        pytest.param("surface_temperature", id="surface_temperature"),
        pytest.param("sst", id="sst"),
        pytest.param("TS", id="TS"),
    ],
)
def test_paired_regional_index_aggregator(variable_name):
    n_lat = 10
    n_lon = 20
    n_sample = 2
    n_times = 365 * 4  # one year of data for monthly averaging
    target_data = _get_windowed_data(
        n_sample, n_times, n_lat, n_lon, sst_name=variable_name
    )
    prediction_data = _get_windowed_data(
        n_sample, n_times, n_lat, n_lon, time_varying=False, sst_name=variable_name
    )
    time = _get_windowed_times((2000, 1, 1, 0, 0, 0), n_sample, n_times)
    lat_lon_coordinates = LatLonCoordinates(
        lat=target_data["lat"],
        lon=target_data["lon"],
    )
    region = LatLonRegion(
        lat=lat_lon_coordinates.lat,
        lon=lat_lon_coordinates.lon,
        lat_bounds=(4.5, 6.5),
        lon_bounds=(9.5, 11.5),
    )
    agg = PairedRegionalIndexAggregator(
        target_aggregator=RegionalIndexAggregator(
            regional_weights=region.regional_weights,
            regional_mean=lat_lon_coordinates.get_gridded_operations().regional_area_weighted_mean,
        ),
        prediction_aggregator=RegionalIndexAggregator(
            regional_weights=region.regional_weights,
            regional_mean=lat_lon_coordinates.get_gridded_operations().regional_area_weighted_mean,
        ),
    )
    agg.record_batch(time=time, target_data=target_data, gen_data=prediction_data)
    logs = agg.get_logs(label="test")
    assert len(logs) > 0
    metric_name = f"test/{variable_name}_nino34_index"
    assert metric_name in logs
    assert isinstance(logs[metric_name], plt.Figure)

    metric_name = f"test/{variable_name}_nino34_index_power_spectrum"
    assert metric_name in logs
    assert isinstance(logs[metric_name], plt.Figure)
